# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-09-15 14:50:00
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-02-10 17:54:00

from __future__ import print_function, division, absolute_import

import distutils
import warnings

from astropy.io import fits
from astropy.wcs import WCS
import numpy as np

import marvin
import marvin.core.exceptions
import marvin.tools.spaxel
import marvin.utils.general.general
import marvin.tools.maps

from marvin.core.core import MarvinToolsClass
from marvin.core.exceptions import MarvinError, MarvinUserWarning


class ModelCube(MarvinToolsClass):
    """A class to interface with MaNGA DAP model cubes.

    This class represents a DAP model cube, initialised either from a file,
    a database, or remotely via the Marvin API.

    Parameters:
        data (``HDUList``, SQLAlchemy object, or None):
            An astropy ``HDUList`` or a SQLAlchemy object of a model cube, to
            be used for initialisation. If ``None``, the normal mode will
            be used (see :ref:`mode-decision-tree`).
        cube (:class:`~marvin.tools.cube.Cube` object)
            The DRP cube object associated with this model cube.
        maps (:class:`~marvin.tools.maps.Maps` object)
            The DAP maps object associated with this model cube. Must match
            the ``bintype``, ``template_kin``, and ``template_pop``.
        filename (str):
            The path of the file containing the model cube to load.
        mangaid (str):
            The mangaid of the model cube to load.
        plateifu (str):
            The plate-ifu of the model cube to load (either ``mangaid`` or
            ``plateifu`` can be used, but not both).
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See :ref:`mode-decision-tree`.
        bintype (str or None):
            The binning type. For MPL-4, one of the following: ``'NONE',
            'RADIAL', 'STON'`` (if ``None`` defaults to ``'NONE'``).
            For MPL-5 and successive, one of, ``'ALL', 'NRE', 'SPX', 'VOR10'``
            (defaults to ``'ALL'``).
        template_kin (str or None):
            The template use for kinematics. For MPL-4, one of
            ``'M11-STELIB-ZSOL', 'MILES-THIN', 'MIUSCAT-THIN'`` (if ``None``,
            defaults to ``'MIUSCAT-THIN'``). For MPL-5 and successive, the only
            option in ``'GAU-MILESHC'`` (``None`` defaults to it).
        template_pop (str or None):
            A placeholder for a future version in which stellar populations
            are fitted using a different template that ``template_kin``. It
            has no effect for now.
        nsa_source ({'auto', 'drpall', 'nsa'}):
            Defines how the NSA data for this object should loaded when
            ``ModelCube.nsa`` is first called. If ``drpall``, the drpall file
            will be used (note that this will only contain a subset of all the
            NSA information); if ``nsa``, the full set of data from the DB will
            be retrieved. If the drpall file or a database are not available, a
            remote API call will be attempted. If ``nsa_source='auto'``, the
            source will depend on how the ``ModelCube`` object has been
            instantiated. If the cube has ``ModelCube.data_origin='file'``,
            the drpall file will be used (as it is more likely that the user
            has that file in their system). Otherwise, ``nsa_source='nsa'``
            will be assumed. This behaviour can be modified during runtime by
            modifying the ``ModelCube.nsa_mode`` with one of the valid values.
        release (str):
            The MPL/DR version of the data to use.

    Return:
        modelcube:
            An object representing the model cube.

    """

    def __init__(self, *args, **kwargs):

        valid_kwargs = [
            'data', 'cube', 'maps', 'filename', 'mangaid', 'plateifu', 'mode',
            'release', 'bintype', 'template_kin', 'template_pop', 'nsa_source']

        assert len(args) == 0, 'Maps does not accept arguments, only keywords.'
        for kw in kwargs:
            assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

        if kwargs.pop('template_pop', None):
            warnings.warn('template_pop is not yet in use. Ignoring value.', MarvinUserWarning)

        self.bintype = kwargs.pop('bintype', marvin.tools.maps.__BINTYPES_UNBINNED__)
        self.template_kin = kwargs.pop('template_kin', marvin.tools.maps.__TEMPLATES_KIN_DEFAULT__)
        self.template_pop = None

        super(ModelCube, self).__init__(*args, **kwargs)

        # Checks that DAP is at least MPL-5
        MPL5 = distutils.version.StrictVersion('2.0.2')
        if self.filename is None and distutils.version.StrictVersion(self._dapver) < MPL5:
            raise MarvinError('ModelCube requires at least dapver=\'2.0.2\'')

        self._cube = kwargs.pop('cube', None)
        self._maps = kwargs.pop('maps', None)

        assert self.bintype in marvin.tools.maps.__BINTYPES__, \
            'bintype must be on of {0}'.format(marvin.tools.maps.__BINTYPES__)
        assert self.template_kin in marvin.tools.maps.__TEMPLATES_KIN__, \
            'template_kin must be on of {0}'.format(marvin.tools.maps.__TEMPLATES_KIN__)

        self.header = None
        self.wcs = None
        self.shape = None
        self.wavelength = None

        if self.data_origin == 'file':
            self._load_modelcube_from_file()
        elif self.data_origin == 'db':
            self._load_modelcube_from_db()
        elif self.data_origin == 'api':
            self._load_modelcube_from_api()
        else:
            raise MarvinError('data_origin={0} is not valid'.format(self.data_origin))

        # Confirm that drpver and dapver match the ones from the header.
        marvin.tools.maps.Maps._check_versions(self)

    def __repr__(self):
        """Representation for ModelCube."""

        return ('<Marvin ModelCube (plateifu={0}, mode={1}, data_origin={2}, bintype={3}, '
                'template_kin={4})>'
                .format(repr(self.plateifu), repr(self.mode),
                        repr(self.data_origin), repr(self.bintype), repr(self.template_kin)))

    def __getitem__(self, xy):
        """Returns the spaxel for ``(x, y)``"""

        return self.getSpaxel(x=xy[1], y=xy[0], xyorig='lower')

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')
        daptype = '{0}-{1}'.format(self.bintype, self.template_kin)

        return super(ModelCube, self)._getFullPath('mangadap5', ifu=ifu,
                                                   drpver=self._drpver,
                                                   dapver=self._dapver,
                                                   plate=plate, mode='LOGCUBE',
                                                   daptype=daptype)

    def download(self):
        """Downloads the cube using sdss_access - Rsync"""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')
        daptype = '{0}-{1}'.format(self.bintype, self.template_kin)

        return super(ModelCube, self).download('mangadap5', ifu=ifu,
                                               drpver=self._drpver,
                                               dapver=self._dapver,
                                               plate=plate, mode='LOGCUBE',
                                               daptype=daptype)

    def _load_modelcube_from_file(self):
        """Initialises a model cube from a file."""

        if self.data is not None:
            assert isinstance(self.data, fits.HDUList), 'data is not an HDUList object'
        else:
            try:
                self.data = fits.open(self.filename)
            except IOError as err:
                raise IOError('filename {0} cannot be found: {1}'.format(self.filename, err))

        self.header = self.data[0].header
        self.shape = self.data['FLUX'].data.shape[1:]
        self.wcs = WCS(self.data['FLUX'].header)
        self.wavelength = self.data['WAVE'].data
        self.redcorr = self.data['REDCORR'].data

        self.plateifu = self.header['PLATEIFU']
        self.mangaid = self.header['MANGAID']

        # Checks and populates release.
        file_drpver = self.header['VERSDRP3']
        file_drpver = 'v1_5_1' if file_drpver == 'v1_5_0' else file_drpver

        file_ver = marvin.config.lookUpRelease(file_drpver)
        assert file_ver is not None, 'cannot find file version.'

        if file_ver != self._release:
            warnings.warn('mismatch between file version={0} and object release={1}. '
                          'Setting object release to {0}'.format(file_ver, self._release),
                          marvin.core.exceptions.MarvinUserWarning)
            self._release = file_ver

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

    def _load_modelcube_from_db(self):
        """Initialises a model cube from the DB."""

        mdb = marvin.marvindb
        plate, ifu = self.plateifu.split('-')

        if not mdb.isdbconnected:
            raise MarvinError('No db connected')

        else:

            datadb = mdb.datadb
            dapdb = mdb.dapdb

            if self.data:
                assert isinstance(self.data, dapdb.ModelCube), \
                    'data is not an instance of marvindb.dapdb.ModelCube.'
            else:
                # Initial query for version
                version_query = mdb.session.query(dapdb.ModelCube).join(
                    dapdb.File,
                    datadb.PipelineInfo,
                    datadb.PipelineVersion).filter(
                        datadb.PipelineVersion.version == self._dapver).from_self()

                # Query for model cube parameters
                db_modelcube = version_query.join(
                    dapdb.File,
                    datadb.Cube,
                    datadb.IFUDesign).filter(
                        datadb.Cube.plate == plate,
                        datadb.IFUDesign.name == str(ifu)).from_self().join(
                            dapdb.File,
                            dapdb.FileType).filter(dapdb.FileType.value == 'LOGCUBE').join(
                                dapdb.Structure, dapdb.BinType).join(
                                    dapdb.Template,
                                    dapdb.Structure.template_kin_pk == dapdb.Template.pk).filter(
                                        dapdb.BinType.name == self.bintype,
                                        dapdb.Template.name == self.template_kin).all()

                if len(db_modelcube) > 1:
                    raise MarvinError('more than one ModelCube found for '
                                      'this combination of parameters.')

                elif len(db_modelcube) == 0:
                    raise MarvinError('no ModelCube found for this combination of parameters.')

                self.data = db_modelcube[0]

            self.header = self.data.file.primary_header
            self.shape = self.data.file.cube.shape.shape
            self.wcs = WCS(self.data.file.cube.wcs.makeHeader())
            self.wavelength = np.array(self.data.file.cube.wavelength.wavelength, dtype=np.float)
            self.redcorr = np.array(self.data.redcorr[0].value, dtype=np.float)

            self.plateifu = str(self.header['PLATEIFU'].strip())
            self.mangaid = str(self.header['MANGAID'].strip())

    def _load_modelcube_from_api(self):
        """Initialises a model cube from the API."""

        url = marvin.config.urlmap['api']['getModelCube']['url']
        url_full = url.format(name=self.plateifu, bintype=self.bintype,
                              template_kin=self.template_kin)

        try:
            response = self._toolInteraction(url_full)
        except Exception as ee:
            raise MarvinError('found a problem when checking if remote model cube '
                              'exists: {0}'.format(str(ee)))

        data = response.getData()

        self.header = fits.Header.fromstring(data['header'])
        self.shape = tuple(data['shape'])
        self.wcs = WCS(fits.Header.fromstring(data['wcs_header']))
        self.wavelength = np.array(data['wavelength'])
        self.redcorr = np.array(data['redcorr'])

        self.bintype = data['bintype']
        self.template_kin = data['template_kin']

        self.plateifu = str(self.header['PLATEIFU'].strip())
        self.mangaid = str(self.header['MANGAID'].strip())

    def getSpaxel(self, x=None, y=None, ra=None, dec=None,
                  spectrum=True, properties=True, **kwargs):
        """Returns the |spaxel| matching certain coordinates.

        The coordinates of the spaxel to return can be input as ``x, y`` pixels
        relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
        coordinates.

        If ``spectrum=True``, the returned |spaxel| will be instantiated with the
        DRP spectrum of the spaxel for the DRP cube associated with this
        ModelCube. The same is true for ``properties=True`` for the DAP
        properties of the spaxel in the Maps associated with these coordinates.

        Parameters:
            x,y (int or array):
                The spaxel coordinates relative to ``xyorig``. If ``x`` is an
                array of coordinates, the size of ``x`` must much that of
                ``y``.
            ra,dec (float or array):
                The coordinates of the spaxel to return. The closest spaxel to
                those coordinates will be returned. If ``ra`` is an array of
                coordinates, the size of ``ra`` must much that of ``dec``.
            xyorig ({'center', 'lower'}):
                The reference point from which ``x`` and ``y`` are measured.
                Valid values are ``'center'`` (default), for the centre of the
                spatial dimensions of the cube, or ``'lower'`` for the
                lower-left corner. This keyword is ignored if ``ra`` and
                ``dec`` are defined.
            spectrum (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding DRP spectrum.
            properties (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding DAP properties for this spaxel.
            modelcube (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding ModelCube data.

        Returns:
            spaxels (list):
                The |spaxel| objects for this cube/maps corresponding to the
                input coordinates. The length of the list is equal to the
                number of input coordinates.

        .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

        """

        kwargs['cube'] = self.cube if spectrum else False
        kwargs['maps'] = self.maps.get_unbinned() if properties else False
        kwargs['modelcube'] = self.get_unbinned()

        return marvin.utils.general.general.getSpaxel(x=x, y=y, ra=ra, dec=dec, **kwargs)

    def _return_extension(self, extension):

        if self.data_origin == 'file':
            return self.data[extension.upper()].data
        elif self.data_origin == 'db':
            return self.data.get3DCube(extension.lower())
        elif self.data_origin == 'api':
            raise MarvinError('cannot return a full cube in remote mode. '
                              'Please use getSpaxel.')

    @property
    def flux(self):
        """Returns the flux extension."""

        return self._return_extension('flux')

    @property
    def ivar(self):
        """Returns the ivar extension."""

        return self._return_extension('ivar')

    @property
    def mask(self):
        """Returns the mask extension."""

        return self._return_extension('mask')

    @property
    def model(self):
        """Returns the model extension."""

        return self._return_extension('model')

    @property
    def emline(self):
        """Returns the emline extension."""

        return self._return_extension('emline')

    @property
    def emline_base(self):
        """Returns the emline_base extension."""

        return self._return_extension('emline_base')

    @property
    def emline_mask(self):
        """Returns the emline_mask extension."""

        return self._return_extension('emline_mask')

    @property
    def stellar_continuum(self):
        """Returns the stellar continuum cube."""

        return (self._return_extension('model') -
                self._return_extension('emline') -
                self._return_extension('emline_base'))

    @property
    def cube(self):
        """Returns the :class:`~marvin.tools.cube.Cube` associated with this ModelCube."""

        if not self._cube:
            if self.data_origin == 'db':
                cube_data = self.data.file.cube
            else:
                cube_data = None

            self._cube = marvin.tools.cube.Cube(data=cube_data,
                                                plateifu=self.plateifu,
                                                release=self._release)

        return self._cube

    @property
    def maps(self):
        """Returns the :class:`~marvin.tools.mas.Maps` associated with this ModelCube."""

        if not self._maps:
            self._maps = marvin.tools.maps.Maps(plateifu=self.plateifu,
                                                bintype=self.bintype,
                                                template_kin=self.template_kin,
                                                release=self._release)

        return self._maps

    def is_binned(self):
        """Returns True if the ModelCube is not unbinned."""

        if marvin.tools.maps._is_MPL4(self._dapver):
            return self.bintype != marvin.tools.maps.__BINTYPES_MPL4_UNBINNED__
        else:
            return self.bintype != marvin.tools.maps.__BINTYPES_UNBINNED__

    def get_unbinned(self):
        """Returns a version of ``self`` corresponding to the unbinned ModelCube."""

        if marvin.tools.maps._is_MPL4(self._dapver):
            unbinned_name = marvin.tools.maps.__BINTYPES_MPL4_UNBINNED__
        else:
            unbinned_name = marvin.tools.maps.__BINTYPES_UNBINNED__

        if self.bintype == unbinned_name:
            return self
        else:
            return ModelCube(plateifu=self.plateifu, release=self._release, bintype=unbinned_name,
                             template_kin=self.template_kin, template_pop=self.template_pop,
                             mode=self.mode)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-05-19 17:56:30
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-06 14:43:00

from __future__ import print_function, division, absolute_import
from marvin.tests.api.conftest import ApiPage
import pytest


@pytest.mark.parametrize('page', [('api', 'getSpectrum')], ids=['getspectrum'], indirect=True)
class TestGetSpectrum(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('x, y', [(17, 17)])
    def test_spec_success(self, galaxy, page, params, reqtype, x, y):
        params.update({'name': galaxy.plateifu, 'x': x, 'y': y})
        data = {'flux': [], 'ivar': [], 'mask': [], 'wavelength': [], 'specres': []}
        page.load_page(reqtype, page.url.format(**params), params=params)
        page.assert_success(data, keys=True)

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg, x, y',
                             [(None, 'release', 'Missing data for required field.', 0, 0),
                              ('badname', 'name', 'String does not match expected pattern.', 0, 0),
                              ('84', 'name', 'Shorter than minimum length 4.', 0, 0),
                              ('8485-1901', 'x', 'Must be between 0 and 100.', -1, 17),
                              ('8485-1901', 'y', 'Must be between 0 and 100.', 17, -1)],
                             ids=['norelease', 'badname', 'shortname', 'badx', 'bady'])
    def test_spec_failure(self, galaxy, page, reqtype, params, name, missing, errmsg, x, y):
        params.update({'name': name, 'x': x, 'y': y})
        if name is None:
            params.update({'name': galaxy.plateifu})
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, params=params, errmsg=errmsg)


@pytest.mark.parametrize('page', [('api', 'getProperties')], ids=['getproperties'], indirect=True)
class TestGetProperties(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('expprop', [('emline_gflux_ha_6564')], ids=['haflux'])
    def test_props_success(self, galaxy, page, params, reqtype, expprop):
        params.update({'name': galaxy.plateifu, 'x': galaxy.dap['x'], 'y': galaxy.dap['y'], 'template_kin': galaxy.template})
        page.load_page(reqtype, page.url.format(**params), params=params)
        page.assert_success()
        expval = galaxy.dap[galaxy.template][expprop]
        props = page.json['data']['properties']
        assert expprop in props
        assert expval == props[expprop]

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg, x, y, template',
                             [(None, 'release', 'Missing data for required field.', 0, 0, None),
                              ('badname', 'name', 'String does not match expected pattern.', 0, 0, None),
                              ('84', 'name', 'Shorter than minimum length 4.', 0, 0, 'GAU-MILESHC'),
                              ('8485-1901', 'x', 'Must be between 0 and 100.', -1, 17, 'GAU-MILESHC'),
                              ('8485-1901', 'y', 'Must be between 0 and 100.', 17, -1, 'GAU-MILESHC'),
                              ('8485-1901', 'template_kin', 'Not a valid choice.', 17, 17, 'MILESHC')],
                             ids=['norelease', 'badname', 'shortname', 'badx', 'bady', 'badtemplate'])
    def test_props_failure(self, galaxy, page, reqtype, params, name, missing, errmsg, x, y, template):
        params.update({'name': name, 'x': x, 'y': y, 'template_kin': template})
        if name is None:
            params.update({'name': galaxy.plateifu})
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, params=params, errmsg=errmsg)


@pytest.mark.parametrize('page', [('api', 'getModels')], ids=['getmodels'], indirect=True)
class TestGetModels(object):

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    def test_models_success(self, galaxy, page, params, reqtype):
        if galaxy.release == 'MPL-4':
            pytest.skip('MPL-4 does not have modelcubes')
        params.update({'name': galaxy.plateifu, 'x': galaxy.dap['x'], 'y': galaxy.dap['y'], 'template_kin': galaxy.template})
        page.load_page(reqtype, page.url.format(**params), params=params)
        data = {'bintype': galaxy.bintype, 'template_kin': galaxy.template, 'flux_array': [],
                'flux_mask': [], 'flux_ivar': [], 'model_array': [], 'model_emline': [],
                'model_emline_base': [], 'model_emline_mask': []}
        page.assert_success(data, keys=True)
        jdata = page.json['data']
        expdata = galaxy.dap[galaxy.template]['model']
        mcdata = [jdata['flux_array'][0], jdata['flux_ivar'][0], jdata['flux_mask'][0],
                  jdata['model_array'][0], jdata['model_emline'][0], jdata['model_emline_base'][0],
                  jdata['model_emline_mask'][0]]
        assert expdata == mcdata

    @pytest.mark.parametrize('reqtype', [('get'), ('post')])
    @pytest.mark.parametrize('name, missing, errmsg, x, y, template',
                             [(None, 'release', 'Missing data for required field.', 0, 0, None),
                              ('badname', 'name', 'String does not match expected pattern.', 0, 0, None),
                              ('84', 'name', 'Shorter than minimum length 4.', 0, 0, 'GAU-MILESHC'),
                              ('8485-1901', 'x', 'Must be between 0 and 100.', -1, 17, 'GAU-MILESHC'),
                              ('8485-1901', 'y', 'Must be between 0 and 100.', 17, -1, 'GAU-MILESHC'),
                              ('8485-1901', 'template_kin', 'Not a valid choice.', 17, 17, 'MILESHC')],
                             ids=['norelease', 'badname', 'shortname', 'badx', 'bady', 'badtemplate'])
    def test_models_failure(self, galaxy, page, reqtype, params, name, missing, errmsg, x, y, template):
        params.update({'name': name, 'x': x, 'y': y, 'template_kin': template})
        if name is None:
            params.update({'name': galaxy.plateifu})
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, errmsg=errmsg)
        else:
            page.route_no_valid_params(page.url.format(**params), missing, reqtype=reqtype, params=params, errmsg=errmsg)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 18:18:54
# @Last modified by:   andrews
# @Last modified time: 2017-07-31 12:07:88

from __future__ import print_function, division, absolute_import
from marvin.tools.query import Query, Results
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.spaxel import Spaxel
from marvin.tools.modelcube import ModelCube
from marvin import config
from marvin.core.exceptions import MarvinError
from collections import OrderedDict, namedtuple
import pytest


myplateifu = '8485-1901'
cols = ['cube.mangaid', 'cube.plate', 'cube.plateifu', 'ifu.name', 'nsa.z']
remotecols = [u'mangaid', u'plate', u'plateifu', u'name', u'z']
coltoparam = OrderedDict([('mangaid', 'cube.mangaid'),
                          ('plate', 'cube.plate'),
                          ('plateifu', 'cube.plateifu'),
                          ('name', 'ifu.name'),
                          ('z', 'nsa.z')])
paramtocol = OrderedDict([('cube.mangaid', 'mangaid'),
                          ('cube.plate', 'plate'),
                          ('cube.plateifu', 'plateifu'),
                          ('ifu.name', 'name'),
                          ('nsa.z', 'z')])


# @pytest.fixture()
# def data(release):

#     rdata = {'MPL-5': {'resdict': {'1': (u'1-209151', 8485, u'8485-12702', u'12702', 0.0185246),
#                                    '2': (u'1-209191', 8485, u'8485-12701', u'12701', 0.0234253),
#                                    '3': (u'1-209113', 8485, u'8485-1902', u'1902', 0.0378877),
#                                    '4': (u'1-209232', 8485, u'8485-1901', u'1901', 0.0407447)},
#                        'row': (u'1-209232', 8485, u'8485-1901', u'1901', 0.0407447),
#                        'count': 4
#                        },
#              'MPL-4': {'resdict': {'1': (u'1-43148', 8135, u'8135-6101', u'6101', 0.0108501),
#                                    '10': (u'1-167079', 8459, u'8459-1901', u'1901', 0.015711),
#                                    '11': (u'1-167075', 8459, u'8459-12704', u'12704', 0.0158584),
#                                    '21': (u'1-113567', 7815, u'7815-12701', u'12701', 0.0167432),
#                                    '31': (u'1-322048', 8552, u'8552-12705', u'12705', 0.0172298),
#                                    '36': (u'1-252151', 8335, u'8335-9102', u'9102', 0.0174864),
#                                    '41': (u'1-378182', 8134, u'8134-12705', u'12705', 0.0178659),
#                                    '46': (u'1-252126', 8335, u'8335-3703', u'3703', 0.0181555)},
#                        'row': (u'1-209232', 8485, u'8485-1901', u'1901', 0.0407447),
#                        'count': 1213
#                        }
#              }

#     return rdata[release]


@pytest.fixture()
def limits(results):
    data = results.expdata['queries'][results.searchfilter]
    count = 10 if data['count'] >= 10 else data['count']
    return (10, count)


@pytest.fixture()
def results(query, request):
    searchfilter = request.param if hasattr(request, 'param') else 'nsa.z < 0.1 and cube.plate==8485'
    q = Query(searchfilter=searchfilter, mode=query.mode, limit=10, release=query._release)
    r = q.run()
    r.expdata = query.expdata
    yield r
    r = None


@pytest.fixture()
def columns(results):
    ''' returns the local or remote column syntax '''
    if results.mode == 'local':
        return cols
    elif results.mode == 'remote':
        return remotecols


class TestResultsColumns(object):

    def test_check_cols(self, results, columns):
        assert set(results.columns) == set(columns)
        assert results.coltoparam == coltoparam
        assert results.paramtocol == paramtocol

    @pytest.mark.parametrize('results', [('nsa.z < 0.1 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('colmns, pars', [(['spaxelprop.x', 'spaxelprop.y', 'emline_gflux_ha_6564', 'bintype.name', 'template.name'],
                                               ['x', 'y', 'emline_gflux_ha_6564', 'bintype_name', 'template_name'])])
    def test_check_withadded(self, results, colmns, pars, columns):
        newcols = cols + colmns
        assert set(results._params) == set(newcols)
        newcols = columns + colmns
        newrem = columns + pars
        if 'name' in newrem:
            newrem[newrem.index('name')] = 'ifu_name'
        assert set(results.columns) == set(newcols) or set(results.columns) == set(newrem)


class TestResultsGetParams(object):

    def test_get_attribute(self, results, columns):
        res = results.results[0]
        for i, name in enumerate(columns):
            assert res[i] == res.__getattribute__(name)

    def test_get_refname(self, results, columns):
        for i, name in enumerate(columns):
            assert name == results._getRefName(cols[i])
            assert name == results._getRefName(remotecols[i])

    @pytest.mark.parametrize('col',
                             [(c) for c in cols] + [(c) for c in remotecols],
                             ids=['cube.mangaid', 'cube.plate', 'cube.plateifu', 'ifu.name', 'nsa.z', u'mangaid', u'plate', u'plateifu', u'name', u'z'])
    def test_get_list(self, results, col):
        obj = results.getListOf(col)
        assert obj is not None
        assert isinstance(obj, list) is True

    @pytest.mark.parametrize('ftype', [('dictlist'), ('listdict')])
    @pytest.mark.parametrize('name', [(None), ('cube.mangaid'), ('nsa.z')], ids=['noname', 'mangaid', 'nsa.z'])
    def test_get_dict(self, results, ftype, name):
        output = results.getDictOf(name, format_type=ftype)

        if ftype == 'listdict':
            assert isinstance(output, list) is True
            assert isinstance(output[0], dict) is True
            if name is not None:
                assert set([name]) == set(list(output[0]))
            else:
                assert set(cols) == set(output[0])
        elif ftype == 'dictlist':
            assert isinstance(output, dict) is True
            assert isinstance(list(output.values())[0], list) is True
            if name is not None:
                assert set([name]) == set(list(output.keys()))
            else:
                assert set(cols) == set(output)


class TestResultsSort(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_sort(self, results, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        results.sort('z')
        limit, count = limits
        data = results.expdata['queries'][results.searchfilter]['sorted']
        assert tuple(data['1']) == results.results[0]
        assert tuple(data[str(count)]) == results.results[count - 1]


class TestResultsPaging(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_check_counts(self, results, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        results.sort('z')
        limit, count = limits
        data = results.expdata['queries'][results.searchfilter]
        assert results.totalcount == data['count']
        assert results.count == count
        assert len(results.results) == count
        assert results.limit == limit
        assert results.chunk == limit

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('chunk, rows',
                             [(10, None),
                              (20, (10, 21))],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_next(self, results, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getNext(chunk=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        else:
            assert results.results[0] == tuple(data['11'])
            assert len(results.results) == chunk
            if rows:
                assert results.results[rows[0]] == tuple(data[str(rows[1])])

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('index, chunk, rows',
                             [(30, 10, [(0, 21)]),
                              (45, 20, [(5, 31), (10, 36), (15, 41)])],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_prev(self, results, index, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getSubset(index, limit=chunk)
        results.getPrevious(chunk=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        elif results.count == 0:
            assert len(results.results) == 0
        else:
            assert len(results.results) == chunk
            if rows:
                for row in rows:
                    assert results.results[row[0]] == tuple(data[str(row[1])])

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('index, chunk, rows',
                             [(35, 10, [(0, 36)]),
                              (30, 20, [(0, 31), (15, 46)])],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_set(self, results, index, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getSubset(index, limit=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        elif results.count == 0:
            assert len(results.results) == 0
        else:
            assert len(results.results) == chunk
            if rows:
                for row in rows:
                    assert results.results[row[0]] == tuple(data[str(row[1])])


class TestResultsPickling(object):

    def test_pickle_save(self, results, temp_scratch):
        file = temp_scratch.join('test_results.mpf')
        path = results.save(str(file), overwrite=True)
        assert file.check() is True

    def test_pickle_restore(self, results, temp_scratch):
        file = temp_scratch.join('test_results.mpf')
        path = results.save(str(file), overwrite=True)
        assert file.check() is True
        r = Results.restore(str(file))
        assert r.searchfilter == results.searchfilter


class TestResultsConvertTool(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('objtype, tool',
                             [('cube', Cube), ('maps', Maps), ('spaxel', Spaxel),
                              ('modelcube', ModelCube)])
    def test_convert_success(self, results, objtype, tool, exporigin):
        if config.release == 'MPL-4' and objtype == 'modelcube':
            pytest.skip('no modelcubes in mpl-4')

        results.convertToTool(objtype, limit=1)
        assert results.objects is not None
        assert isinstance(results.objects, list) is True
        assert isinstance(results.objects[0], tool) is True
        if objtype != 'spaxel':
            assert results.mode == results.objects[0].mode

    @pytest.mark.parametrize('objtype, error, errmsg',
                             [('modelcube', MarvinError, "ModelCube requires at least dapver='2.0.2'"),
                              ('spaxel', AssertionError, 'Parameters must include spaxelprop.x and y in order to convert to Marvin Spaxel')],
                             ids=['mcminrelease', 'nospaxinfo'])
    def test_convert_failures(self, results, objtype, error, errmsg):
        if config.release > 'MPL-4' and objtype == 'modelcube':
            pytest.skip('modelcubes in post mpl-4')

        with pytest.raises(error) as cm:
            results.convertToTool(objtype, limit=1)
        assert cm.type == error
        assert errmsg in str(cm.value)





# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 18:41:25
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-12 18:58:55

from __future__ import print_function, division, absolute_import
import pytest


@pytest.fixture()
def nsa_target(maindb, galaxy):
    nsa = maindb.session.query(maindb.sampledb.NSA).join(maindb.sampledb.MangaTargetToNSA,
                                                         maindb.sampledb.MangaTarget).\
        filter(maindb.sampledb.MangaTarget.mangaid == galaxy.mangaid).one()
    yield nsa
    nsa = None


class TestSampleDB(object):

    @pytest.mark.parametrize('band,expected', [('u', 18.69765903),
                                               ('g', 17.45450578),
                                               ('r', 16.80842176),
                                               ('i', 16.43652498),
                                               ('z', 16.20534984)])
    def test_elpetro_mag(self, nsa_target, band, expected):

        assert pytest.approx(getattr(nsa_target, 'elpetro_mag_{0}'.format(band)), expected)

    @pytest.mark.parametrize('bands,expected', [(('u', 'g'), 1.24315324),
                                                (('g', 'r'), 0.64608403),
                                                (('r', 'i'), 0.37189678),
                                                (('i', 'z'), 0.23117514)])
    def test_elpetro_colour(self, nsa_target, bands, expected):

        bandA, bandB = bands
        assert pytest.approx(nsa_target.elpetro_colour(bandA, bandB), expected)
        elpetro_mag_colour = getattr(nsa_target,
                                     'elpetro_mag_{0}_{1}'.format(bandA, bandB))
        assert pytest.approx(elpetro_mag_colour, expected)

    @pytest.mark.parametrize('bands,expected', [(('u', 'g'), 1.1655902862549006),
                                                (('g', 'r'), 0.5961246490479013),
                                                (('r', 'i'), 0.3375816345214986),
                                                (('i', 'z'), 0.20068740844720168)])
    def test_elpetro_absmag_colour(self, nsa_target, bands, expected):

        bandA, bandB = bands
        assert pytest.approx(nsa_target.elpetro_absmag_colour(bandA, bandB), expected)
        elpetro_absmag_colour = getattr(nsa_target, 'elpetro_absmag_{0}_{1}'.format(bandA, bandB))
        assert pytest.approx(elpetro_absmag_colour, expected)




#!/usr/bin/env python
# encoding: utf-8

'''
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-17 17:46:57
    Last Modified On: 2016-02-17 17:46:57 by Brian

'''
from __future__ import print_function
from __future__ import division
from brain.api.base import BrainBaseView
from brain.api.general import BrainGeneralRequestsView
from marvin import config
from marvin.api import ArgValidator
import json

arg_validate = ArgValidator(urlmap=config.urlmap)


class BaseView(BrainBaseView):
    '''Super Class for all API Views to handle all global API items of interest'''

    def add_config(self):
        utahconfig = {'utahconfig': {'mode': config.mode, 'release': config.release}}
        self.update_results(utahconfig)

    def _pop_args(self, kwargs, arglist=None):
        if arglist:
            arglist = [arglist] if not isinstance(arglist, (list, tuple)) else arglist

            for item in arglist:
                tmp = kwargs.pop(item, None)

        return kwargs

    def before_request(self, *args, **kwargs):
        super(BaseView, self).before_request(*args, **kwargs)

        # try to get a local version of the urlmap for the arg_validator
        if not arg_validate.urlmap:
            bv = BrainGeneralRequestsView()
            resp = bv.buildRouteMap()
            config.urlmap = json.loads(resp.get_data())['urlmap']
            arg_validate.urlmap = config.urlmap




from __future__ import print_function
from marvin.core.exceptions import MarvinError, MarvinUserWarning, MarvinBreadCrumb
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.rss import RSS
from marvin.tools.modelcube import ModelCube
from marvin.tools.spaxel import Spaxel
from marvin import config, log
from marvin.utils.general import getImagesByList, downloadList
from marvin.api.api import Interaction
import warnings
import json
import os
import datetime
from functools import wraps
from astropy.table import Table
from collections import OrderedDict, namedtuple
from marvin.core import marvin_pickle

try:
    import cPickle as pickle
except:
    import pickle

try:
    import pandas as pd
except ImportError:
    warnings.warn('Could not import pandas.', MarvinUserWarning)

__all__ = ['Results']

breadcrumb = MarvinBreadCrumb()


def local_mode_only(fxn):
    '''Decorator that bypasses function if in remote mode.'''

    @wraps(fxn)
    def wrapper(self, *args, **kwargs):
        if self.mode == 'remote':
            raise MarvinError('{0} not available in remote mode'.format(fxn.__name__))
        else:
            return fxn(self, *args, **kwargs)
    return wrapper


def remote_mode_only(fxn):
    '''Decorator that bypasses function if in local mode.'''

    @wraps(fxn)
    def wrapper(self, *args, **kwargs):
        if self.mode == 'local':
            raise MarvinError('{0} not available in local mode'.format(fxn.__name__))
        else:
            return fxn(self, *args, **kwargs)
    return wrapper


class Results(object):
    ''' A class to handle results from queries on the MaNGA dataset

        Parameters:
            results (list):
                List of results satisfying the input Query
            query (object / str):
                The query used to produce these results. In local mode, the query is an
                SQLalchemy object that can be used to redo the query, or extract subsets
                of results from the query. In remote more, the query is a literal string
                representation of the SQL query.
            returntype (str):
                The MarvinTools object to convert the results into.  If initially set, the results
                are automaticaly converted into the specified Marvin Tool Object on initialization
            objects (list):
                The list of Marvin Tools objects created by returntype
            count (int):
                The number of objects in the returned query results
            totalcount (int):
                The total number of objects in the full query results
            mode ({'auto', 'local', 'remote'}):
                The load mode to use. See :doc:`Mode secision tree</mode_decision>`.
            chunk (int):
                For paginated results, the number of results to return.  Defaults to 10.
            start (int):
                For paginated results, the starting index value of the results.  Defaults to 0.
            end (int):
                For paginated results, the ending index value of the resutls.  Defaults to start+chunk.

        Returns:
            results: An object representing the Results entity

        Example:
            >>> f = 'nsa.z < 0.012 and ifu.name = 19*'
            >>> q = Query(searchfilter=f)
            >>> r = q.run()
            >>> print(r)
            >>> Results(results=[(u'4-3602', u'1902', -9999.0), (u'4-3862', u'1902', -9999.0), (u'4-3293', u'1901', -9999.0), (u'4-3988', u'1901', -9999.0), (u'4-4602', u'1901', -9999.0)],
            >>>         query=<sqlalchemy.orm.query.Query object at 0x115217090>,
            >>>         count=64,
            >>>         mode=local)

    '''

    def __init__(self, *args, **kwargs):

        self.results = kwargs.get('results', None)
        self._queryobj = kwargs.get('queryobj', None)
        self._updateQueryObj(**kwargs)
        self.count = kwargs.get('count', None)
        self.totalcount = kwargs.get('totalcount', self.count)
        self._runtime = kwargs.get('runtime', None)
        self.query_runtime = self._getRunTime() if self._runtime is not None else None
        self.mode = config.mode if not kwargs.get('mode', None) else kwargs.get('mode', None)
        self.chunk = self.limit if self.limit else kwargs.get('chunk', 100)
        self.start = kwargs.get('start', 0)
        self.end = kwargs.get('end', self.start + self.chunk)
        self.coltoparam = None
        self.paramtocol = None
        self.objects = None
        self.sortcol = None
        self.order = None

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinResults {0}'.format(self.__class__),
                        category=self.__class__)

        # Convert remote results to NamedTuple
        if self.mode == 'remote':
            self._makeNamedTuple()

        # Setup columns, and parameters
        if self.count > 0:
            self._setupColumns()

        # Auto convert to Marvin Object
        if self.returntype:
            self.convertToTool(self.returntype)

    def _updateQueryObj(self, **kwargs):
        ''' update parameters using the _queryobj '''
        self.query = self._queryobj.query if self._queryobj else kwargs.get('query', None)
        self.returntype = self._queryobj.returntype if self._queryobj else kwargs.get('returntype', None)
        self.searchfilter = self._queryobj.searchfilter if self._queryobj else kwargs.get('searchfilter', None)
        self.returnparams = self._queryobj._returnparams if self._queryobj else kwargs.get('returnparams', None)
        self.limit = self._queryobj.limit if self._queryobj else kwargs.get('limit', None)
        self._params = self._queryobj.params if self._queryobj else kwargs.get('params', None)

    def __repr__(self):
        return ('Marvin Results(query={0}, \ntotalcount={1}, \ncount={2}, \nmode={3}'.format(repr(self.query), self.totalcount, self.count, self.mode))

    def showQuery(self):
        ''' Displays the literal SQL query used to generate the Results objects

            Returns:
                querystring (str):
                    A string representation of the SQL query
        '''

        # check unicode or str
        try:
            isstr = type(self.query) == unicode
        except NameError as e:
            isstr = type(self.query) == str

        # return the string query or compile the real query
        if isstr:
            return self.query
        else:
            return str(self.query.statement.compile(compile_kwargs={'literal_binds': True}))

    def _getRunTime(self):
        ''' Sets the query runtime as a datetime timedelta object '''
        if type(self._runtime) == dict:
            return datetime.timedelta(**self._runtime)
        else:
            return self._runtime

    def download(self, images=False, limit=None):
        ''' Download results via sdss_access

            Uses sdss_access to download the query results via rsync.
            Downloads them to the local sas. The data type downloaded
            is indicated by the returntype parameter

            i.e. $SAS_BASE_DIR/mangawork/manga/spectro/redux/...

            Parameters:
                images (bool):
                    Set to only download the images of the query results

            Returns:
                NA: na

            Example:
                >>> r = q.run()
                >>> r.returntype = 'cube'
                >>> r.download()
        '''

        plates = self.getListOf(name='plate')
        ifus = self.getListOf(name='ifu.name')
        plateifu = ['{0}-{1}'.format(z[0], z[1]) for z in zip(plates, ifus)]
        if images:
            tmp = getImagesByList(plateifu, mode='remote', as_url=True, download=True)
        else:
            downloadList(plateifu, dltype=self.returntype, limit=limit)

    def sort(self, name, order='asc'):
        ''' Sort the set of results by column name

            Sorts the results by a given parameter / column name.  Sets
            the results to the new sorted results.

            Parameters:
                name (str):
                order ({'asc', 'desc'}):

            Returns:
                sortedres (list):
                    The listed of sorted results.

            Example:
                >>> r = q.run()
                >>> r.getColumns()
                >>> [u'mangaid', u'name', u'nsa.z']
                >>> r.results
                >>> [(u'4-3988', u'1901', -9999.0),
                >>>  (u'4-3862', u'1902', -9999.0),
                >>>  (u'4-3293', u'1901', -9999.0),
                >>>  (u'4-3602', u'1902', -9999.0),
                >>>  (u'4-4602', u'1901', -9999.0)]

                >>> # Sort the results by mangaid
                >>> r.sort('mangaid')
                >>> [(u'4-3293', u'1901', -9999.0),
                >>>  (u'4-3602', u'1902', -9999.0),
                >>>  (u'4-3862', u'1902', -9999.0),
                >>>  (u'4-3988', u'1901', -9999.0),
                >>>  (u'4-4602', u'1901', -9999.0)]

                >>> # Sort the results by IFU name in descending order
                >>> r.sort('ifu.name', order='desc')
                >>> [(u'4-3602', u'1902', -9999.0),
                >>>  (u'4-3862', u'1902', -9999.0),
                >>>  (u'4-3293', u'1901', -9999.0),
                >>>  (u'4-3988', u'1901', -9999.0),
                >>>  (u'4-4602', u'1901', -9999.0)]
        '''
        refname = self._getRefName(name)
        self.sortcol = refname
        self.order = order

        if self.mode == 'local':
            reverse = True if order == 'desc' else False
            sortedres = sorted(self.results, key=lambda row: row.__getattribute__(refname), reverse=reverse)
            self.results = sortedres
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['querycubes']['url']

            params = {'searchfilter': self.searchfilter, 'params': self.returnparams,
                      'sort': refname, 'order': order, 'limit': self.limit}
            try:
                ii = Interaction(route=url, params=params)
            except MarvinError as e:
                raise MarvinError('API Query Sort call failed: {0}'.format(e))
            else:
                self.results = ii.getData()
                self._makeNamedTuple()
                sortedres = self.results

        return sortedres

    def toTable(self):
        ''' Output the results as an Astropy Table

            Uses the Python Astropy package

            Parameters:
                None

            Returns:
                tableres:
                    Astropy Table

            Example:
                >>> r = q.run()
                >>> r.toTable()
                >>> <Table length=5>
                >>> mangaid    name   nsa.z
                >>> unicode6 unicode4   float64
                >>> -------- -------- ------------
                >>>   4-3602     1902      -9999.0
                >>>   4-3862     1902      -9999.0
                >>>   4-3293     1901      -9999.0
                >>>   4-3988     1901      -9999.0
                >>>   4-4602     1901      -9999.0
        '''
        try:
            tabres = Table(rows=self.results, names=self.getColumns())
        except ValueError as e:
            raise MarvinError('Could not make astropy Table from results: {0}'.format(e))
        return tabres

    def toFits(self, filename='myresults.fits'):
        ''' Output the results as a FITS file

        Writes a new FITS file from search results using
        the astropy Table.write()

        Parameters:
            filename (str):
                Name of FITS file to output

        '''
        myext = os.path.splitext(filename)[1]
        if not myext:
            filename = filename+'.fits'
        table = self.toTable()
        table.write(filename, format='fits')
        print('Writing new FITS file {0}'.format(filename))

    def toCSV(self, filename='myresults.csv'):
        ''' Output the results as a CSV file

        Writes a new CSV file from search results using
        the astropy Table.write()

        Parameters:
            filename (str):
                Name of CSV file to output

        '''
        myext = os.path.splitext(filename)[1]
        if not myext:
            filename = filename+'.csv'
        table = self.toTable()
        table.write(filename, format='csv')
        print('Writing new CSV file {0}'.format(filename))

    def toDF(self):
        '''Call toDataFrame().
        '''
        return self.toDataFrame()

    def toDataFrame(self):
        '''Output the results as an pandas dataframe.

            Uses the pandas package.

            Parameters:
                None

            Returns:
                dfres:
                    pandas dataframe

            Example:
                >>> r = q.run()
                >>> r.toDataFrame()
                mangaid  plate   name     nsa_mstar         z
                0  1-22286   7992  12704  1.702470e+11  0.099954
                1  1-22301   7992   6101  9.369260e+10  0.105153
                2  1-22414   7992   6103  7.489660e+10  0.092272
                3  1-22942   7992  12705  8.470360e+10  0.104958
                4  1-22948   7992   9102  1.023530e+11  0.119399
        '''
        try:
            dfres = pd.DataFrame(self.results)
        except (ValueError, NameError) as e:
            raise MarvinError('Could not make pandas dataframe from results: {0}'.format(e))
        return dfres

    def _makeNamedTuple(self):
        ''' '''
        ntnames = self._reduceNames(self._params, under=True)
        try:
            nt = namedtuple('NamedTuple', ntnames)
        except ValueError as e:
            raise MarvinError('Cannot created NamedTuple from remote Results: {0}'.format(e))
        else:
            globals()[nt.__name__] = nt

        qpo = ntnames

        def keys(self):
            return qpo
        nt.keys = keys
        self.results = [nt(*r) for r in self.results]

    def save(self, path=None, overwrite=False):
        ''' Save the results as a pickle object

        Parameters:
            path (str):
                Filepath and name of the pickled object
            overwrite (bool):
                Set this to overwrite an existing pickled file

        Returns:
            path (str):
                The filepath and name of the pickled object

        '''

        # set Marvin query object to None, in theory this could be pickled as well
        self._queryobj = None
        try:
            isnotstr = type(self.query) != unicode
        except NameError as e:
            isnotstr = type(self.query) != str
        if isnotstr:
            self.query = None

        sf = self.searchfilter.replace(' ', '') if self.searchfilter else 'anon'
        # set the path
        if not path:
            path = os.path.expanduser('~/marvin_results_{0}.mpf'.format(sf))

        # check for file extension
        if not os.path.splitext(path)[1]:
            path = os.path.join(path+'.mpf')

        path = os.path.realpath(path)

        if os.path.isdir(path):
            raise MarvinError('path must be a full route, including the filename.')

        if os.path.exists(path) and not overwrite:
            warnings.warn('file already exists. Not overwriting.', MarvinUserWarning)
            return

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        try:
            pickle.dump(self, open(path, 'wb'), protocol=-1)
        except Exception as ee:
            if os.path.exists(path):
                os.remove(path)
            raise MarvinError('Error found while pickling: {0}'.format(str(ee)))

        return path

    @classmethod
    def restore(cls, path, delete=False):
        ''' Restore a pickled Results object

        Parameters:
            path (str):
                The filename and path to the pickled object

            delete (bool):
                Turn this on to delete the pickled fil upon restore

        Returns:
            Results (instance):
                The instantiated Marvin Results class
        '''
        return marvin_pickle.restore(path, delete=delete)

    def toJson(self):
        ''' Output the results as a JSON object

            Uses Python json package to convert the results to JSON representation

            Parameters:
                None

            Returns:
                jsonres:
                    JSONed results

            Example:
                >>> r = q.run()
                >>> r.toJson()
                >>> '[["4-3602", "1902", -9999.0], ["4-3862", "1902", -9999.0], ["4-3293", "1901", -9999.0],
                >>>   ["4-3988", "1901", -9999.0], ["4-4602", "1901", -9999.0]]'
        '''
        try:
            jsonres = json.dumps(self.results)
        except TypeError as e:
            raise MarvinError('Results not JSON-ifiable. Check the format of results: {0}'.format(e))
        return jsonres

    def getColumns(self):
        ''' Get the column names of the returned reults

            Parameters:
                None

            Returns:
                columns (list):
                    A list of column names from the results

            Example:
                >>> r = q.run()
                >>> cols = r.getColumns()
                >>> print(cols)
                >>> [u'mangaid', u'name', u'nsa.z']
        '''
        try:
            self.columns = list(self.results[0].keys()) if self.results else None
        except Exception as e:
            raise MarvinError('Could not get table keys from results.  Results not an SQLalchemy results collection: {0}'.format(e))
        return self.columns

    def _reduceNames(self, columns, under=None):
        ''' reduces the full table.parameter names to non-dotted versions '''
        # fullstr = ','.join(columns)
        # names = [t if '.' not in t else t.split('.')[1] if fullstr.count(
        #             '.'+t.split('.')[1]) == 1 else t for t in columns]
        colsplit = [c if '.' not in c else c.split('.')[1] for c in columns]
        names = [t if '.' not in t else t.split('.')[1] if colsplit.count(
                    t.split('.')[1]) == 1 else t for t in columns]
        # replace . with _
        if under:
            names = [n if '.' not in n else n.replace('.', '_')for n in names]
        return names

    def _setupColumns(self):
        ''' Auto sets up all the column/parameter name info '''
        columns = self.getColumns()
        redcol = self._reduceNames(columns)
        tmp = self.mapColumnsToParams(inputs=redcol)
        tmp = self.mapParamsToColumns(inputs=redcol)

    def mapColumnsToParams(self, col=None, inputs=None):
        ''' Map the columns names from results to the original parameter names '''
        columns = self.getColumns()
        #params = self._params if self.mode == 'local' else self._params
        cols = columns if not inputs else inputs
        if cols:
            if not self.coltoparam:
                self.coltoparam = OrderedDict(zip(cols, self._params))
            mapping = self.coltoparam[col] if col else list(self.coltoparam.values())
        else:
            mapping = None
        return mapping

    def mapParamsToColumns(self, param=None, inputs=None):
        ''' Map original parameter names to the column names '''
        columns = self.getColumns()
        #params = self._params if self.mode == 'local' else self._params
        cols = columns if not inputs else inputs
        if cols:
            if not self.paramtocol:
                self.paramtocol = OrderedDict(zip(self._params, cols))
            mapping = self.paramtocol[param] if param else list(self.paramtocol.values())
        else:
            mapping = None
        return mapping

    def getListOf(self, name=None, to_json=False):
        ''' Extract a list of a single parameter from results

            Parameters:
                name (str):
                    Name of the parameter name to return.  If not specified,
                    it returns all parameters.
                to_json (bool):
                    True/False boolean to convert the output into a JSON format

            Returns:
                output (list):
                    A list of results for one parameter

            Example:
                >>> r = q.run()
                >>> r.getListOf('mangaid')
                >>> [u'4-3988', u'4-3862', u'4-3293', u'4-3602', u'4-4602']

            Raises:
                AssertionError:
                    Raised when no name is specified.
        '''

        assert name, 'Must specify a column name'

        # get reference name
        refname = self._getRefName(name)
        if not refname:
            raise MarvinError('Name {0} not a property in results.  Try another.'.format(refname))

        output = None
        try:
            output = [r.__getattribute__(refname) for r in self.results]
        except AttributeError as e:
            raise MarvinError('Name {0} not a property in results.  Try another: {1}'.format(refname, e))

        if to_json:
            output = json.dumps(output) if output else None

        return output

    def getDictOf(self, name=None, format_type='listdict', to_json=False):
        ''' Get a dictionary of specified parameters

            Parameters:
                name (str):
                    Name of the parameter name to return.  If not specified,
                    it returns all parameters.
                format_type ({'listdict', 'dictlist'}):
                    The format of the results. Listdict is a list of dictionaries.
                    Dictlist is a dictionary of lists. Default is listdict.
                to_json (bool):
                    True/False boolean to convert the output into a JSON format

            Returns:
                output (list, dict):
                    Can be either a list of dictionaries, or a dictionary of lists

            Example:
                >>> # get some results
                >>> r = q.run()
                >>> # Get a list of dictionaries
                >>> r.getDictOf(format_type='listdict')
                >>> [{'cube.mangaid': u'4-3988', 'ifu.name': u'1901', 'nsa.z': -9999.0},
                >>>  {'cube.mangaid': u'4-3862', 'ifu.name': u'1902', 'nsa.z': -9999.0},
                >>>  {'cube.mangaid': u'4-3293', 'ifu.name': u'1901', 'nsa.z': -9999.0},
                >>>  {'cube.mangaid': u'4-3602', 'ifu.name': u'1902', 'nsa.z': -9999.0},
                >>>  {'cube.mangaid': u'4-4602', 'ifu.name': u'1901', 'nsa.z': -9999.0}]

                >>> # Get a dictionary of lists
                >>> r.getDictOf(format_type='dictlist')
                >>> {'cube.mangaid': [u'4-3988', u'4-3862', u'4-3293', u'4-3602', u'4-4602'],
                >>>  'ifu.name': [u'1901', u'1902', u'1901', u'1902', u'1901'],
                >>>  'nsa.z': [-9999.0, -9999.0, -9999.0, -9999.0, -9999.0]}

                >>> # Get a dictionary of only one parameter
                >>> r.getDictOf('mangaid')
                >>> [{'cube.mangaid': u'4-3988'},
                >>>  {'cube.mangaid': u'4-3862'},
                >>>  {'cube.mangaid': u'4-3293'},
                >>>  {'cube.mangaid': u'4-3602'},
                >>>  {'cube.mangaid': u'4-4602'}]

        '''

        # Try to get the sqlalchemy results keys
        keys = self.getColumns()
        tmp = self.mapColumnsToParams()
        tmp = self.mapParamsToColumns()

        if self.mode == 'local':
            lookup = OrderedDict(zip(keys, keys))
        elif self.mode == 'remote':
            lookup = self.coltoparam

        # Format results
        if format_type == 'listdict':
            output = [{lookup[k]: res.__getattribute__(k) for k in keys} for res in self.results]
        elif format_type == 'dictlist':
            output = {lookup[k]: [res.__getattribute__(k) for res in self.results] for k in keys}
        else:
            raise MarvinError('No output.  Check your input format_type.')

        # Test if name is in results
        if name:
            refname = self._getRefName(name)
            print('refname', refname)
            if refname:
                # Format results
                newname = lookup[refname]
                if format_type == 'listdict':
                    output = [{newname: i[newname]} for i in output]
                elif format_type == 'dictlist':
                    output = {newname: output[newname]}
                else:
                    output = None
            else:
                raise MarvinError('Name {0} not a property in results.  Try another'.format(name))

        if to_json:
            output = json.dumps(output) if output else None

        return output

    def _getRefName(self, name):
        ''' Get the appropriate reference column / parameter name

        This converts an input name into the respective column name
        '''
        cols = self.getColumns()

        if name in cols:
            # name already in the list of column names
            refname = name
        else:
            if not self.coltoparam:
                pars = self.mapColumnsToParams()
            if not self.paramtocol:
                tmp = self.mapParamsToColumns()

            refname = self.coltoparam[name] if name in self.coltoparam else \
                self.paramtocol[name] if name in self.paramtocol else None

        return refname

    def getNext(self, chunk=None):
        ''' Retrieve the next chunk of results

            Returns the next chunk of results from the query.
            from start to end in units of chunk.  Used with getPrevious
            to paginate through a long list of results

            Parameters:
                chunk (int):
                    The number of objects to return

            Returns:
                results (list):
                    A list of query results

            Example:
                >>> r = q.run()
                >>> r.getNext(5)
                >>> Retrieving next 5, from 35 to 40
                >>> [(u'4-4231', u'1902', -9999.0),
                >>>  (u'4-14340', u'1901', -9999.0),
                >>>  (u'4-14510', u'1902', -9999.0),
                >>>  (u'4-13634', u'1901', -9999.0),
                >>>  (u'4-13538', u'1902', -9999.0)]

        '''

        if chunk and chunk < 0:
            warnings.warn('Chunk cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            chunk = self.chunk

        newstart = self.end
        self.chunk = chunk if chunk else self.chunk
        newend = newstart + self.chunk

        # This handles cases when the number of results is < total
        if self.totalcount == self.count:
            warnings.warn('You have all the results.  Cannot go forward', MarvinUserWarning)
            return self.results

        # This handles the end edge case
        if newend > self.totalcount:
            warnings.warn('You have reached the end.', MarvinUserWarning)
            newend = self.totalcount
            newstart = self.end

        # This grabs the next chunk
        log.info('Retrieving next {0}, from {1} to {2}'.format(self.chunk, newstart, newend))
        if self.mode == 'local':
            self.results = self.query.slice(newstart, newend).all()
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']

            params = {'searchfilter': self.searchfilter, 'params': self.returnparams,
                      'start': newstart, 'end': newend, 'limit': self.limit,
                      'sort': self.sortcol, 'order': self.order}
            try:
                ii = Interaction(route=url, params=params)
            except MarvinError as e:
                raise MarvinError('API Query GetNext call failed: {0}'.format(e))
            else:
                self.results = ii.getData()
                self._makeNamedTuple()

        self.start = newstart
        self.end = newend
        self.count = len(self.results)

        if self.returntype:
            self.convertToTool()

        return self.results

    def getPrevious(self, chunk=None):
        ''' Retrieve the previous chunk of results.

            Returns a previous chunk of results from the query.
            from start to end in units of chunk.  Used with getNext
            to paginate through a long list of results

            Parameters:
                chunk (int):
                    The number of objects to return

            Returns:
                results (list):
                    A list of query results

            Example:
                >>> r = q.run()
                >>> r.getPrevious(5)
                >>> Retrieving previous 5, from 30 to 35
                >>> [(u'4-3988', u'1901', -9999.0),
                >>>  (u'4-3862', u'1902', -9999.0),
                >>>  (u'4-3293', u'1901', -9999.0),
                >>>  (u'4-3602', u'1902', -9999.0),
                >>>  (u'4-4602', u'1901', -9999.0)]

         '''

        if chunk and chunk < 0:
            warnings.warn('Chunk cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            chunk = self.chunk

        newend = self.start
        self.chunk = chunk if chunk else self.chunk
        newstart = newend - self.chunk

        # This handles cases when the number of results is < total
        if self.totalcount == self.count:
            warnings.warn('You have all the results.  Cannot go back', MarvinUserWarning)
            return self.results

        # This handles the start edge case
        if newstart < 0:
            warnings.warn('You have reached the beginning.', MarvinUserWarning)
            newstart = 0
            newend = self.start

        # This grabs the previous chunk
        log.info('Retrieving previous {0}, from {1} to {2}'.format(self.chunk, newstart, newend))
        if self.mode == 'local':
            self.results = self.query.slice(newstart, newend).all()
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']

            params = {'searchfilter': self.searchfilter, 'params': self.returnparams,
                      'start': newstart, 'end': newend, 'limit': self.limit,
                      'sort': self.sortcol, 'order': self.order}
            try:
                ii = Interaction(route=url, params=params)
            except MarvinError as e:
                raise MarvinError('API Query GetNext call failed: {0}'.format(e))
            else:
                self.results = ii.getData()
                self._makeNamedTuple()

        self.start = newstart
        self.end = newend
        self.count = len(self.results)

        if self.returntype:
            self.convertToTool()

        return self.results

    def getSubset(self, start, limit=10):
        ''' Extracts a subset of results

            Parameters:
                start (int):
                    The starting index of your subset extraction
                limit (int):
                    The limiting number of results to return.

            Returns:
                results (list):
                    A list of query results

            Example:
                >>> r = q.run()
                >>> r.getSubset(0, 10)
                >>> [(u'14-12', u'1901', -9999.0),
                >>> (u'14-13', u'1902', -9999.0),
                >>> (u'27-134', u'1901', -9999.0),
                >>> (u'27-100', u'1902', -9999.0),
                >>> (u'27-762', u'1901', -9999.0),
                >>> (u'27-759', u'1902', -9999.0),
                >>> (u'27-827', u'1901', -9999.0),
                >>> (u'27-828', u'1902', -9999.0),
                >>> (u'27-1170', u'1901', -9999.0),
                >>> (u'27-1167', u'1902', -9999.0)]
        '''

        if limit < 0:
            warnings.warn('Limit cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            limit = self.chunk

        start = 0 if int(start) < 0 else int(start)
        end = start + int(limit)
        # if end > self.count:
        #     end = self.count
        #     start = end - int(limit)
        self.start = start
        self.end = end
        self.chunk = limit
        if self.mode == 'local':
            self.results = self.query.slice(start, end).all()
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']

            params = {'searchfilter': self.searchfilter, 'params': self.returnparams,
                      'start': start, 'end': end, 'limit': self.limit,
                      'sort': self.sortcol, 'order': self.order}
            try:
                ii = Interaction(route=url, params=params)
            except MarvinError as e:
                raise MarvinError('API Query GetNext call failed: {0}'.format(e))
            else:
                self.results = ii.getData()
                self._makeNamedTuple()

        self.count = len(self.results)
        if self.returntype:
            self.convertToTool()

        return self.results

    @local_mode_only
    def getAll(self):
        ''' Retrieve all of the results of a query

            Parameters:
                None

            Returns:
                results (list):
                    A list of query results.
        '''
        self.results = self.query.all()
        return self.results

    def convertToTool(self, tooltype, **kwargs):
        ''' Converts the list of results into Marvin Tool objects

            Creates a list of Marvin Tool objects from a set of query results.
            The new list is stored in the Results.objects property.
            If the Query.returntype parameter is specified, then the Results object
            will automatically convert the results to the desired Tool on initialization.

            Parameters:
                tooltype (str):
                    The requested Marvin Tool object that the results are converted into.
                    Overrides the returntype parameter.  If not set, defaults
                    to the returntype parameter.
                limit (int):
                    Limit the number of results you convert to Marvin tools.  Useful
                    for extremely large result sets.  Default is None.
                mode (str):
                    The mode to use when attempting to convert to Tool. Default mode
                    is to use the mode internal to Results. (most often remote mode)

            Example:
                >>> # Get the results from some query
                >>> r = q.run()
                >>> r.results
                >>> [NamedTuple(mangaid=u'14-12', name=u'1901', nsa.z=-9999.0),
                >>>  NamedTuple(mangaid=u'14-13', name=u'1902', nsa.z=-9999.0),
                >>>  NamedTuple(mangaid=u'27-134', name=u'1901', nsa.z=-9999.0),
                >>>  NamedTuple(mangaid=u'27-100', name=u'1902', nsa.z=-9999.0),
                >>>  NamedTuple(mangaid=u'27-762', name=u'1901', nsa.z=-9999.0)]

                >>> # convert results to Marvin Cube tools
                >>> r.convertToTool('cube')
                >>> r.objects
                >>> [<Marvin Cube (plateifu='7444-1901', mode='remote', data_origin='api')>,
                >>>  <Marvin Cube (plateifu='7444-1902', mode='remote', data_origin='api')>,
                >>>  <Marvin Cube (plateifu='7995-1901', mode='remote', data_origin='api')>,
                >>>  <Marvin Cube (plateifu='7995-1902', mode='remote', data_origin='api')>,
                >>>  <Marvin Cube (plateifu='8000-1901', mode='remote', data_origin='api')>]

        '''

        # set the desired tool type
        mode = kwargs.get('mode', self.mode)
        limit = kwargs.get('limit', None)
        toollist = ['cube', 'spaxel', 'maps', 'rss', 'modelcube']
        tooltype = tooltype if tooltype else self.returntype
        assert tooltype in toollist, 'Returned tool type must be one of {0}'.format(toollist)

        # get the parameter list to check against
        paramlist = self.paramtocol.keys()

        print('Converting results to Marvin {0} objects'.format(tooltype.title()))
        if tooltype == 'cube':
            self.objects = [Cube(plateifu=res.__getattribute__(
                            self._getRefName('cube.plateifu')),
                            mode=mode) for res in self.results[0:limit]]
        elif tooltype == 'maps':

            isbin = 'bintype.name' in paramlist
            istemp = 'template.name' in paramlist
            self.objects = []

            for res in self.results[0:limit]:
                mapkwargs = {'mode': mode, 'plateifu': res.__getattribute__(self._getRefName('cube.plateifu'))}

                if isbin:
                    binval = res.__getattribute__(self._getRefName('bintype.name'))
                    mapkwargs['bintype'] = binval

                if istemp:
                    tempval = res.__getattribute__(self._getRefName('template.name'))
                    mapkwargs['template_kin'] = tempval

                self.objects.append(Maps(**mapkwargs))
        elif tooltype == 'spaxel':

            assert 'spaxelprop.x' in paramlist and 'spaxelprop.y' in paramlist, \
                    'Parameters must include spaxelprop.x and y in order to convert to Marvin Spaxel.'

            self.objects = []

            load = kwargs.get('load', True)
            maps = kwargs.get('maps', True)
            modelcube = kwargs.get('modelcube', True)

            for res in self.results[0:limit]:
                spaxkwargs = {'plateifu': res.__getattribute__(self._getRefName('cube.plateifu')),
                              'x': res.__getattribute__(self._getRefName('spaxelprop.x')),
                              'y': res.__getattribute__(self._getRefName('spaxelprop.y')),
                              'load': load, 'maps': maps, 'modelcube': modelcube}
                self.objects.append(Spaxel(**spaxkwargs))
        elif tooltype == 'rss':
            self.objects = [RSS(plateifu=res.__getattribute__(
                            self._getRefName('cube.plateifu')),
                            mode=mode) for res in self.results[0:limit]]
        elif tooltype == 'modelcube':

            isbin = 'bintype.name' in paramlist
            istemp = 'template.name' in paramlist
            self.objects = []

            for res in self.results[0:limit]:
                mapkwargs = {'mode': mode, 'plateifu': res.__getattribute__(self._getRefName('cube.plateifu'))}

                if isbin:
                    binval = res.__getattribute__(self._getRefName('bintype.name'))
                    mapkwargs['bintype'] = binval

                if istemp:
                    tempval = res.__getattribute__(self._getRefName('template.name'))
                    mapkwargs['template_kin'] = tempval

                self.objects.append(ModelCube(**mapkwargs))




from __future__ import absolute_import, division, print_function, unicode_literals
# The line above will help with 2to3 support.

import sys
try:
    from urllib2 import HTTPPasswordMgrWithDefaultRealm, HTTPBasicAuthHandler, build_opener, install_opener, urlopen
    from urllib2 import HTTPError
except:
    from urllib.request import HTTPPasswordMgrWithDefaultRealm, HTTPBasicAuthHandler, build_opener, install_opener, urlopen
    from urllib.error import HTTPError

from os import makedirs
from os.path import isfile, exists, dirname
from sdss_access import SDSSPath
from sdss_access.sync.auth import Auth

class HttpAccess(SDSSPath):
    """Class for providing HTTP access via urllib.request (python3) or urllib2 (python2) to SDSS SAS Paths
    """

    def __init__(self, verbose=False):
        super(HttpAccess, self).__init__()
        self.verbose = verbose
        self._remote = False

    def set_auth(self, username=None, password=None):
        self.auth = Auth(netloc=self.netloc,verbose=self.verbose)
        self.auth.set_username(username)
        self.auth.set_password(password)
        if not self.auth.ready(): self.auth.load()

    def remote(self, remote_base=None, username=None, password=None):
        """
        Configures remote access

        Parameters
        ----------
        remote_base : str
            base URL path for remote repository
        username : str
            user name for remote repository
        password : str
            password for local repository
        """
        if remote_base is not None: self.remote_base = remote_base
        self._remote = True
        self.set_auth(username=username,password=password)
        if self.auth.ready():
            passman = HTTPPasswordMgrWithDefaultRealm()
            passman.add_password(None, self.remote_base, self.auth.username, self.auth.password)
            authhandler = HTTPBasicAuthHandler(passman)
            opener = build_opener(authhandler)
            install_opener(opener)

    def local(self):
        """Configures back to local access
        """
        self._remote = False

    def get(self, filetype, **kwargs):
        """Returns file name, downloading if remote access configured.

        Parameters
        ----------
        filetype : str
            type of file

        keyword arguments :
            keywords to fully specify path

        Notes
        -----
        Path templates are defined in $DIMAGE_DIR/data/dimage_paths.ini
        """
        
        path = self.full(filetype, **kwargs)
        
        if path:
            if self._remote: self.download_url_to_path(self.url(filetype, **kwargs), path)
        else: print("There is no file with filetype=%r to access in the tree module loaded" % filetype)

    def download_url_to_path(self, url, path, force=False):
        """
        Download a file from url via http, and put it at path

        Parameters
        ----------

        url : str
            URL of file to download

        path : str
            local path to put file in
        """

        path_exists = isfile(path)
        if not path_exists or force:

            dir = dirname(path)
            if not exists(dir):
                if self.verbose: print("CREATE %s" % dir)
                makedirs(dir)

            try: u = urlopen(url)
            except HTTPError as e:
                u = None
                print("HTTP error code %r.  Please check you ~/.netrc has the correct authorization" % e.code)

            if u:
                with open(path, 'wb') as file:
                    meta = u.info()
                    meta_func = meta.getheaders \
                        if hasattr(meta, 'getheaders') else meta.get_all
                    meta_length = meta_func("Content-Length")
                    file_size = None
                    if meta_length: file_size = int(meta_length[0])
                    if self.verbose: print("Downloading: {0} Bytes: {1}".format(url, file_size))

                    file_size_dl = 0
                    block_sz = 8192
                    while True:
                        buffer = u.read(block_sz)
                        if not buffer: break
                        file_size_dl += len(buffer)
                        file.write(buffer)
        
                if self.verbose:
                    if path_exists: print("OVERWRITING %s" % path)
                    else: print("CREATE %s" % path)

        elif self.verbose: print("FOUND %s (already downloaded)" % path)




# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-04-28 11:34:06
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-07-30 19:46:40

from __future__ import print_function, division, absolute_import
import pytest
from marvin.web import create_app
from marvin.web.settings import TestConfig, CustomConfig
from marvin.api.api import Interaction
from marvin import marvindb, config
from marvin.web.extensions import limiter
from flask import template_rendered, templating
from contextlib import contextmanager
import os
import numpy as np
try:
    from urllib.parse import urlparse, urljoin
except ImportError:
    from urlparse import urlparse, urljoin


# @pytest.fixture(scope='session')
# def drpver(release):
#     drpver, dapver = config.lookUpVersions(release)
#     return drpver


# @pytest.fixture(scope='session')
# def dapver(release):
#     drpver, dapver = config.lookUpVersions(release)
#     return dapver


@pytest.fixture(scope='session')
def app():
    object_config = type('Config', (TestConfig, CustomConfig), dict())
    app = create_app(debug=True, local=True, object_config=object_config)
    limiter.enabled = False
    return app


# def set_sasurl(loc='local', port=None):
#     if not port:
#         port = int(os.environ.get('LOCAL_MARVIN_PORT', 5000))
#     istest = True if loc == 'utah' else False
#     config.switchSasUrl(loc, test=istest, port=port)
#     response = Interaction('api/general/getroutemap', request_type='get')
#     config.urlmap = response.getRouteMap()


# @pytest.fixture()
# def saslocal():
#     set_sasurl(loc='local')


def test_db_stuff():
    assert marvindb is not None
    assert marvindb.datadb is not None
    assert marvindb.sampledb is not None
    assert marvindb.dapdb is not None
    assert 'local' == marvindb.dbtype


@pytest.fixture(scope='function')
def init_web(monkeypatch, set_config):
    config.forceDbOn()

    # monkeypath the render templating to nothing
    def _empty_render(template, context, app):
        template_rendered.send(app, template=template, context=context)
        return ""
    monkeypatch.setattr(templating, '_render', _empty_render)


@pytest.fixture(scope='function', autouse=True)
def inspection(monkeypatch):
    from brain.core.inspection import Inspection
    try:
        monkeypatch.setattr('inspection.marvin.Inspection', Inspection)
    except Exception as e:
        pass


@pytest.mark.usefixtures('app, get_templates')
class Page(object):
    ''' Object representing a Web Page '''
    def __init__(self, client, blue, endpoint):
        self.app = app()
        self.url = self.get_url(blue, endpoint)
        self.json = None
        self.data = None
        self.response = None
        self.client = client

    def get_url(self, blue, endpoint):
        return config.urlmap[blue][endpoint]['url']

    def load_page(self, reqtype, page, params=None):
        if reqtype == 'get':
            self.response = self.client.get(page, query_string=params)
        elif reqtype == 'post':
            self.response = self.client.post(page, data=params, content_type='application/x-www-form-urlencoded')
        self.load_data()

    def load_data(self):
        try:
            self.json = self.response.json
        except ValueError as e:
            self.json = None
        self.data = self.json['data'] if self.json and 'data' in self.json else ''

    def assert_webjson_success(self, expdata):
        self.assert200(message='response status should be 200 for ok')
        if isinstance(expdata, str):
            assert expdata in self.json['result']
        elif isinstance(expdata, dict):
            assert self.json['result']['status'] == 1
        elif isinstance(expdata, list):
            self.assertListIn(expdata, self.json['result'])

    def route_no_valid_webparams(self, template, context, noparam, reqtype='get', params=None, errmsg=None):
        self.assert422(message='response status should be 422 for invalid params')
        assert 'errors/unprocessable_entity.html' == template.name, 'template name should be unprocessable_entity'
        noparam = [noparam] if not isinstance(noparam, list) else noparam
        invalid = {p: [errmsg] for p in noparam}
        assert context['data'] == invalid, 'response should contain validation error dictionary'

    # Assert definitions from Flask-Testing
    def assertListIn(self, a, b):
        ''' assert all items in list a are in b '''
        for item in a:
            assert item in b

    @staticmethod
    def _compare_values_is_subset(aa, bb):
        """Checks if one value or list is a subset of other."""

        if not hasattr(aa, '__iter__') and not hasattr(aa, '__getitem__'):
            if aa != bb and not np.isclose(aa, bb):
                return False
        else:
            # Checks whether the elements are a list of lists. If so, recursively calls itself.
            try:
                if not set(aa).issubset(set(bb)):
                    return False
            except Exception:
                if len(aa) > len(bb):
                    return False
                else:
                    for ii in range(len(aa)):
                        return Page._compare_values_is_subset(aa[ii], bb[ii])

        return True

    def assert_dict_contains_subset(self, subset, dictionary):
        """Asserts whether a dictionary is a subset of other."""

        missing = []
        mismatched = []
        for key, value in subset.items():
            if key not in dictionary:
                missing.append(key)
            elif not self._compare_values_is_subset(value, dictionary[key]):
                mismatched.append((key, (value, dictionary[key])))

        assert not (missing or mismatched), \
            '{0} dictionary should be subset of {1}'.format(subset, dictionary)

    def assert_status(self, status_code, message=None):
        message = message or 'HTTP Status {0} expected but got {1}'.format(status_code, self.response.status_code)
        assert self.response.status_code == status_code, message

    def assert200(self, message=None):
        self.assert_status(200, message)

    def assert400(self, message=None):
        self.assert_status(400, message)

    def assert401(self, message=None):
        self.assert_status(401, message)

    def assert403(self, message=None):
        self.assert_status(403, message)

    def assert404(self, message=None):
        self.assert_status(404, message)

    def assert405(self, message=None):
        self.assert_status(405, message)

    def assert422(self, message=None):
        self.assert_status(422, message)

    def assert500(self, message=None):
        self.assert_status(500, message)

    def assert_redirects(self, location, message=None):
        parts = urlparse(location)
        if parts.netloc:
            expected_location = location
        else:
            server_name = self.app.config.get('SERVER_NAME') or 'localhost'
            expected_location = urljoin('http://{0}'.format(server_name), location)

        valid_status_codes = (301, 302, 303, 305, 307)
        valid_status_code_str = ', '.join(str(code) for code in valid_status_codes)
        not_redirect = "HTTP Status {0} expected but got {1}".format(valid_status_code_str, self.response.status_code)
        assert self.response.status_code in valid_status_codes, message or not_redirect
        assert self.response.location == expected_location, message


@pytest.fixture()
def page(client, config, request, init_web):
    blue, endpoint = request.param
    page = Page(client, blue, endpoint)
    yield page


@contextmanager
def captured_templates(app):
    ''' Records which templates are used '''
    recorded = []

    def record(app, template, context, **extra):
        recorded.append((template, context))

    template_rendered.connect(record)
    yield recorded
    template_rendered.disconnect(record)


@pytest.fixture()
def get_templates(app):
    ''' Fixture that returns which jinja template used '''
    with captured_templates(app) as templates:
        yield templates



#!/usr/bin/env python
# encoding: utf-8
#
# Licensed under a 3-clause BSD license.
#
#
# map.py
#
# Created by Brett Andrews on 28 Apr 2017.
#
# Includes code from mangadap.plot.maps.py licensed under the following 3-clause
# BSD license.
#
# Copyright (c) 2015, SDSS-IV/MaNGA Pipeline Group
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from __future__ import division, print_function, absolute_import

import copy

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from marvin import config
import marvin.utils.plot.colorbar as colorbar
from marvin.utils.general import get_plot_params


def no_coverage_mask(mask, bit, ivar=None):
    """Mask spaxels that are not covered by the IFU.

    Parameters:
        mask (array):
            Mask for value.
        bit (int):
            Bit for "NOCOV."
        ivar (array):
            Inverse variance for image. Default is None.

    Returns:
        array: Boolean array for mask (i.e., True corresponds to value to be
        masked out).
    """
    assert (bit is not None) or (ivar is not None), 'Must provide a bit or ivar array.'

    return (mask & 2**bit).astype(bool) if bit is not None else (ivar == 0)


def bad_data_mask(mask, bits):
    """Mask spaxels that are flagged as bad data by the DAP.

    The masks that are considered bad data are "UNRELIABLE" and "DONOTUSE."
    Note: MPL-4 used only a good = 0 and bad = 1 mask. The "bad" flag
    corresponds most closely to "DONOTUSE."

    Parameters:
        mask (array):
            Mask for value.
        bits (dict):
            Bits that indicate bad data.

    Returns:
        array: Boolean array for mask (i.e., True corresponds to value to be
        masked out).
    """
    if 'unreliable' in bits.keys():     
        unreliable = (mask & 2**bits['unreliable']).astype(bool)
    else:
        unreliable = np.zeros(mask.shape, dtype=bool)

    donotuse = (mask & 2**bits['doNotUse']).astype(bool)
    return np.logical_or.reduce((unreliable, donotuse))


def low_snr_mask(value, ivar, snr_min):
    """Mask spaxels with a signal-to-noise ratio below some threshold.

    Parameters:
        value (array):
            Value for image.
        ivar (array):
            Inverse variance of value.
        snr_min (float):
            Minimum signal-to-noise for keeping a valid measurement.

    Returns:
        array: Boolean array for mask (i.e., True corresponds to value to be
        masked out).
    """
    low_snr = np.zeros(value.shape, dtype=bool)

    if (ivar is not None) and (not np.all(np.isnan(ivar))):
        low_snr = (ivar == 0.)

        if snr_min is not None:
            low_snr[np.abs(value * np.sqrt(ivar)) < snr_min] = True

    return low_snr


def log_colorbar_mask(value, log_cb):
    """Mask spaxels with negative value when using logarithmic colorbar.

    Parameters:
        value (array):
            Value for image.
        log_cb (bool):
            Use logarithmic colorbar.

    Returns:
        array: Boolean array for mask (i.e., True corresponds to value to be
        masked out).
    """
    mask = np.zeros(value.shape, dtype=bool)

    if log_cb:
        mask[value <= 0.] = True

    return mask

def _get_prop(title):
    """Gets property name from plot title.
    
    Parameters:
        title (str):
            Plot title.
    
    Returns:
        str
    """
    if 'vel' in title:
        return 'vel'
    elif 'sigma' in title:
        return 'sigma'
    else:
        return 'default'

def select_good_spaxels(value, nocov, bad_data, low_snr, log_cb_mask):
    """Create masked array of spaxels to display.

    Parameters:
        value (array):
            Value for image.
        nocov (array):
            Mask for spaxels without IFU coverage.
        bad_data (array):
            Mask for data flagged as bad (see
            :ref:`marvin-utils-plot-map-default-params` for default bits).
        low_snr (array):
            Mask for data below the signal-to-noise ratio threshold.
        low_cb_mask (array):
            Mask for negative elements of ``value`` if using a logarithmic
            colorbar.

    Returns:
        masked array: spaxels to display in plot.
    """
    return np.ma.array(value, mask=np.logical_or.reduce((nocov, bad_data, low_snr, log_cb_mask)))


def set_extent(cube_size, sky_coords):
    """Set extent of map.

    Parameters:
        cube_size (tuple):
            Size of the cube in spaxels.
        sky_coords (bool):
            If True, use sky coordinates, otherwise use spaxel coordinates.

    Returns:
        array
    """
    if sky_coords:
        spaxel_size = 0.5  # arcsec
        extent = np.array([-(cube_size[0] * spaxel_size), (cube_size[0] * spaxel_size),
                           -(cube_size[1] * spaxel_size), (cube_size[1] * spaxel_size)])
    else:
        extent = np.array([0, cube_size[0] - 1, 0, cube_size[1] - 1])

    return extent


def set_patch_style(extent, facecolor='#A8A8A8'):
    """Set default parameters for a patch.

    Parameters:
        extent (tuple):
            Extent of image (xmin, xmax, ymin, ymax).
        facecolor (str):
            Background color. Default is '#A8A8A8' (gray).

    Returns:
        dict
    """
    patch_kws = dict(xy=(extent[0] + 0.01, extent[2] + 0.01),
                     width=extent[1] - extent[0] - 0.02,
                     height=extent[3] - extent[2] - 0.02, hatch='xxxx', linewidth=0,
                     fill=True, facecolor=facecolor, edgecolor='w', zorder=0)

    return patch_kws


def ax_setup(sky_coords, fig=None, ax=None, facecolor='#A8A8A8'):
    """Do basic axis setup for maps.

    Parameters:
        sky_coords (bool):
            If True, show plot in sky coordinates (i.e., arcsec), otherwise
            show in spaxel coordinates.
        fig (plt.figure object):
            Matplotlib plt.figure object. Use if creating subplot of a
            multi-panel plot. Default is None.
        ax (plt.figure axis object):
            Matplotlib plt.figure axis object. Use if creating subplot of a
            multi-panel plot. Default is None.
        facecolor (str):
            Axis facecolor. Default is '#A8A8A8'.

    Returns:
        tuple: (plt.figure object, plt.figure axis object)
    """
    xlabel = 'arcsec' if sky_coords else 'spaxel'
    ylabel = 'arcsec' if sky_coords else 'spaxel'

    if ax is None:
        fig, ax = plt.subplots()

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if int(mpl.__version__.split('.')[0]) <= 1:
        ax.set_axis_bgcolor(facecolor)
    else:
        ax.set_facecolor(facecolor)
        ax.grid(False, which='both', axis='both')

    return fig, ax


def set_title(title=None, property_name=None, channel=None):
    """Set title for map.

    Parameters:
        title (str):
            If ``None``, try to set automatically from property (and channel)
            name(s). For no title, set to ''. Default is ``None``.
        property_str (str):
            Map property name. Default is ``None``.
        channel (str):
            Map channel name. Default is ``None``.
    Returns:
        str
    """
    if title is None:
        property_name = property_name if property_name is not None else ''
        channel = channel if channel is not None else ''
        title = ' '.join((property_name, channel))
        title = ' '.join(title.split('_')).strip()

    return title


def plot(*args, **kwargs):
    """Make single panel map or one panel of multi-panel map plot.

    Parameters:
        dapmap (marvin.tools.map.Map):
            Marvin Map object. Default is ``None``.
        value (array):
            Data array. Default is ``None``.
        ivar (array):
            Inverse variance array. Default is ``None``.
        mask (array):
            Mask array. Default is ``None``.
        cmap (str):
            Colormap (see :ref:`marvin-utils-plot-map-default-params` for
            defaults).
        percentile_clip (tuple-like):
            Percentile clip (see :ref:`marvin-utils-plot-map-default-params`
            for defaults).
        sigma_clip (float):
            Sigma clip. Default is ``False``.
        cbrange (tuple-like):
            If ``None``, set automatically. Default is ``None``.
        symmetric (bool):
            Draw a colorbar that is symmetric around zero (see
            :ref:`marvin-utils-plot-map-default-params` for default).
        snr_min (float):
            Minimum signal-to-noise for keeping a valid measurement (see
            :ref:`marvin-utils-plot-map-default-params` for default).
        log_cb (bool):
            Draw a log normalized colorbar. Default is ``False``.
        title (str):
            If ``None``, set automatically from property (and channel) name(s).
            For no title, set to ''. Default is ``None``.
        cblabel (str):
            If ``None``, set automatically from unit. For no colorbar label,
            set to ''. Default is ``None``.
        sky_coords (bool):
            If ``True``, show plot in sky coordinates (i.e., arcsec), otherwise
            show in spaxel coordinates. Default is ``False``.
        use_mask (bool):
            Use DAP bitmasks. Default is ``True``.
        plt_style (str):
            Matplotlib style sheet to use. Default is 'seaborn-darkgrid'.
        fig (matplotlib Figure object):
            Use if creating subplot of a multi-panel plot. Default is ``None``.
        ax (matplotlib Axis object):
            Use if creating subplot of a multi-panel plot. Default is ``None``.
        imshow_kws (dict):
            Keyword args to pass to `ax.imshow
            <http://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.imshow.html#matplotlib.axes.Axes.imshow>`_.
            Default is ``None``.
        cb_kws (dict):
            Keyword args to set and draw colorbar. Default is ``None``.
        return_cb (bool):
            Return colorbar axis. Default it ``False``.

    Returns:
        fig, ax (tuple):
            `matplotlib.figure <http://matplotlib.org/api/figure_api.html>`_,
            `matplotlib.axes <http://matplotlib.org/api/axes_api.html>`_

    Example:
        >>> import marvin.utils.plot.map as mapplot
        >>> maps = Maps(plateifu='8485-1901')
        >>> ha = maps['emline_gflux_ha_6564']
        >>> fig, ax = mapplot.plot(dapmap=ha)
    """
    valid_kwargs = ['dapmap', 'value', 'ivar', 'mask', 'cmap', 'percentile_clip', 'sigma_clip',
                    'cbrange', 'symmetric', 'snr_min', 'log_cb', 'title', 'cblabel', 'sky_coords',
                    'use_mask', 'plt_style', 'fig', 'ax', 'imshow_kws', 'cb_kws', 'return_cb']

    assert len(args) == 0, 'Map.plot() does not accept arguments, only keywords.'

    for kw in kwargs:
        assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

    assert ((kwargs.get('percentile_clip', False)) +
            (kwargs.get('sigma_clip', False)) +
            (kwargs.get('cbrange', None) is not None) <= 1), \
        'Only set one of percentile_clip, sigma_clip, or cbrange!'

    dapmap = kwargs.get('dapmap', None)
    value = kwargs.get('value', None)
    ivar = kwargs.get('ivar', None)
    mask = kwargs.get('mask', None)
    sigma_clip = kwargs.get('sigma_clip', False)
    cbrange = kwargs.get('cbrange', None)
    log_cb = kwargs.get('log_cb', False)
    title = kwargs.get('title', None)
    cblabel = kwargs.get('cblabel', None)
    sky_coords = kwargs.get('sky_coords', False)
    use_mask = kwargs.get('use_mask', True)
    plt_style = kwargs.get('plt_style', 'seaborn-darkgrid')
    fig = kwargs.get('fig', None)
    ax = kwargs.get('ax', None)
    imshow_kws = kwargs.get('imshow_kws', {})
    cb_kws = kwargs.get('cb_kws', {})
    return_cb = kwargs.get('return_cb', False)

    assert (value is not None) or (dapmap is not None), \
        'Map.plot() requires specifying ``value`` or ``dapmap``.'

    # user-defined value, ivar, or mask overrides dapmap attributes
    value = value if value is not None else getattr(dapmap, 'value', None)
    ivar = ivar if ivar is not None else getattr(dapmap, 'ivar', None)
    mask = mask if mask is not None else getattr(dapmap, 'mask', np.zeros(value.shape, dtype=bool))

    title = set_title(title,
                      property_name=getattr(dapmap, 'property_name', None),
                      channel=getattr(dapmap, 'channel', None))

    # get plotparams from datamodel
    dapver = config.lookUpVersions()[1]
    prop = _get_prop(title)
    params = get_plot_params(dapver, prop)
    cmap = kwargs.get('cmap', params['cmap'])
    percentile_clip = kwargs.get('percentile_clip', params['percentile_clip'])
    symmetric = kwargs.get('symmetric', params['symmetric'])
    snr_min = kwargs.get('snr_min', params['snr_min'])

    if sigma_clip:
        percentile_clip = False

    # create no coverage, bad data, low SNR, and log colorbar masks
    nocov_mask = no_coverage_mask(mask, params['bitmasks'].get('nocov', None), ivar)
    badData = params['bitmasks']['badData']
    bad_data = bad_data_mask(mask, badData) if use_mask else np.zeros(value.shape)
    low_snr = low_snr_mask(value, ivar, snr_min) if use_mask else np.zeros(value.shape)
    log_cb_mask = log_colorbar_mask(value, log_cb)

    # final masked array to show
    good_spax = select_good_spaxels(value, nocov_mask, bad_data, low_snr, log_cb_mask)

    # setup colorbar
    cb_kws['cmap'] = cmap
    cb_kws['percentile_clip'] = percentile_clip
    cb_kws['sigma_clip'] = sigma_clip
    cb_kws['cbrange'] = cbrange
    cb_kws['symmetric'] = symmetric
    cb_kws['label'] = cblabel if cblabel is not None else getattr(dapmap, 'unit', '')
    cb_kws['log_cb'] = log_cb
    cb_kws = colorbar._set_cb_kws(cb_kws)
    cb_kws = colorbar._set_cbrange(good_spax, cb_kws)

    # setup unmasked spaxels
    extent = set_extent(value.shape, sky_coords)
    imshow_kws.setdefault('extent', extent)
    imshow_kws.setdefault('interpolation', 'nearest')
    imshow_kws.setdefault('origin', 'lower')
    imshow_kws['norm'] = LogNorm() if log_cb else None

    # setup background
    nocov_kws = copy.deepcopy(imshow_kws)
    nocov = np.ma.array(np.ones(value.shape), mask=~nocov_mask)
    A8A8A8 = colorbar._one_color_cmap(color='#A8A8A8')

    # setup masked spaxels
    patch_kws = set_patch_style(extent=extent)

    # finish setup of unmasked spaxels and colorbar range
    imshow_kws = colorbar._set_vmin_vmax(imshow_kws, cb_kws['cbrange'])

    # set hatch color and linewidths (in matplotlib 2.0+)
    try:
        mpl_rc = {it: mpl.rcParams[it] for it in ['hatch.color', 'hatch.linewidth']}
        mpl.rc_context({'hatch.color': 'w', 'hatch.linewidth': '0.5'})
    except KeyError as ee:
        mpl_rc = {}

    with plt.style.context(plt_style):

        fig, ax = ax_setup(sky_coords=sky_coords, fig=fig, ax=ax)

        # plot hatched regions by putting one large patch as lowest layer
        # hatched regions are bad data, low SNR, or negative values if the colorbar is logarithmic
        ax.add_patch(mpl.patches.Rectangle(**patch_kws))

        # plot regions without IFU coverage as a solid color (gray #A8A8A8)
        ax.imshow(nocov, cmap=A8A8A8, zorder=1, **nocov_kws)

        # plot unmasked spaxels
        p = ax.imshow(good_spax, cmap=cb_kws['cmap'], zorder=10, **imshow_kws)

        fig, cb = colorbar._draw_colorbar(fig, mappable=p, ax=ax, **cb_kws)

        if title is not '':
            ax.set_title(label=title)

    # restore previous matplotlib rc parameters (as of matplotlib 2.0.2 this
    # redraws the hatches with the original rcParam settings)
    # mpl.rc_context(mpl_rc)

    # turn on to preserve zorder when saving to pdf (or other vector based graphics format)
    mpl.rcParams['image.composite_image'] = False

    output = (fig, ax) if not return_cb else (fig, ax, cb)
    return output



#!/usr/bin/env python
# encoding: utf-8


# Created by Brian Cherinka on 2016-05-10 20:17:52
# Licensed under a 3-clause BSD license.

# Revision History:
#     Initial Version: 2016-05-10 20:17:52 by Brian Cherinka
#     Last Modified On: 2016-05-10 20:17:52 by Brian


from __future__ import print_function, division
from marvin.core.core import MarvinToolsClass
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from marvin import config
import numpy as np
import warnings
from functools import wraps

try:
    from sdss_access import RsyncAccess, AccessError
    from sdss_access.path import Path
except ImportError:
    Path = None
    RsyncAccess = None

__all__ = ['Image']


class Image(MarvinToolsClass):
    '''A class to interface with MaNGA images.

     - DRP optical images
     - DAP MAP images
     - NSA preimaging
     mangafile = 'optical' or 'map' or 'nsa'

    - maybe Image is new Core base class instead of MarvinToolsClass?
    - and we have new classes for MapImage, DRPImage, NSAImage?

    '''

    def __init__(self, *args, **kwargs):

        self.data_origin = None
        self._drpver = kwargs.get('drpver', config.drpver)
        self._dapver = kwargs.get('dapver', config.dapver)
        self._mplver = kwargs.get('mplver', config.mplver)
        self.download = kwargs.get('download', config.download)

        MarvinToolsClass.__init__(self, *args, **kwargs)

        if self.mode == 'local':
            if self.filename:
                # do any file stuff
                pass
            else:
                # do any db stuff
                self.plate, self.ifu = self.plateifu.split('-')

        else:
            # do any remote stuff
            pass

    def __repr__(self):
        '''Image representation.'''
        return '<Marvin Image (plate={0}, ifu={1}, drpver={2})>'.format(self.plate, self.ifu, self.drpver)

    def getByPlate(self):
        pass

    def getByList(self):
        pass

    def getRandom(self):
        pass





#!/usr/bin/env python
# encoding: utf-8
#
# conftest.py
#
# Created by Brett Andrews on 20 Mar 2017.


from collections import OrderedDict
import itertools
import os

import pytest

from marvin import config, marvindb
from marvin.api.api import Interaction
from marvin.tools.cube import Cube
from marvin.tools.modelcube import ModelCube
from marvin.tools.maps import Maps
from marvin.tools.query import Query
from marvin.tools.maps import _get_bintemps, __BINTYPES_MPL4__, __TEMPLATES_KIN_MPL4__
from sdss_access.path import Path
import yaml


# PYTEST MODIFIERS
# -----------------
def pytest_addoption(parser):
    """Add new options"""
    # run slow tests
    parser.addoption('--runslow', action='store_true', default=False, help='Run slow tests.')
    # control releases run
    parser.addoption('--travis-only', action='store_true', default=False, help='Run a Travis only subset')


def pytest_runtest_setup(item):
    """Skip slow tests."""
    if 'slow' in item.keywords and not item.config.getoption('--runslow'):
        pytest.skip('Requires --runslow option to run.')


def pytest_configure(config):
    ''' Runs during configuration of conftest.  Checks and sets a global instance for a
        TravisSubset based on the pytest command line input of --travis-only
    '''
    option = config.getoption('--travis-only')
    global travis
    if option:
        travis = TravisSubset()

# specific release instance
travis = None


class TravisSubset(object):
    def __init__(self):
        self.new_gals = ['8485-1901']
        self.new_releases = ['MPL-5']
        self.new_bintypes = ['SPX']  # ['SPX', 'VOR10', 'NONE', 'STON']
        self.new_templates = ['GAU-MILESHC', 'MILES-THIN']
        self.new_dbs = ['nodb']
        self.new_origins = ['file', 'api']
        self.new_modes = ['local', 'remote']


# Global Parameters for FIXTURES
# ------------------------------
releases = ['MPL-5', 'MPL-4']           # to loop over releases (see release fixture)


def populate_bintypes_templates(releases):
    ''' Generates bintype and template dictionaries for each release '''
    bintypes = OrderedDict((release, []) for release in releases)
    templates = OrderedDict((release, []) for release in releases)
    for release in releases:
        __, dapver = config.lookUpVersions(release)
        bintemps = _get_bintemps(dapver)
        for bintemp in bintemps:
            bintype = bintemp.split('-')[0]
            template = '-'.join(bintemp.split('-')[1:])
            if bintype not in bintypes[release]:
                bintypes[release].append(bintype)
            if template not in templates[release]:
                templates[release].append(template)
    return bintypes, templates


bintypes, templates = populate_bintypes_templates(releases)

# TODO reduce modes to only local and remote
modes = ['local', 'remote', 'auto']     # to loop over modes (see mode fixture)
dbs = ['db', 'nodb']                    # to loop over dbs (see db fixture)
origins = ['file', 'db', 'api']         # to loop over data origins (see data_origin fixture)


# Galaxy and Query data is stored in a YAML file
galaxy_data = yaml.load(open(os.path.join(os.path.dirname(__file__), 'data/galaxy_test_data.dat')))
query_data = yaml.load(open(os.path.join(os.path.dirname(__file__), 'data/query_test_data.dat')))


@pytest.fixture(scope='session', params=releases)
def release(request):
    """Yield a release."""
    return request.param


def _get_release_generator_chain():
    """Return all valid combinations of (release, bintype, template)."""
    return itertools.chain(*[itertools.product([release], bintypes[release],
                                               templates[release]) for release in releases])


def _params_ids(fixture_value):
    """Return a test id for the release chain."""
    return '-'.join(fixture_value)


@pytest.fixture(scope='session', params=sorted(_get_release_generator_chain()), ids=_params_ids)
def get_params(request):
    """Yield a tuple of (release, bintype, template)."""

    # placeholder until the marvin_test_if decorator works in 2.7
    release, bintype, template = request.param
    if travis and release not in travis.new_releases:
        pytest.skip('Skipping non-requested release')

    if travis and bintype not in travis.new_bintypes:
        pytest.skip('Skipping non-requested release')

    if travis and template not in travis.new_templates:
        pytest.skip('Skipping non-requested release')

    return request.param


@pytest.fixture(scope='session', params=sorted(galaxy_data.keys()))
def plateifu(request):
    """Yield a plate-ifu."""
    if travis and request.param not in travis.new_gals:
        pytest.skip('Skipping non-requested galaxies')
    return request.param


@pytest.fixture(scope='session', params=origins)
def data_origin(request):
    """Yield a data access mode."""
    if travis and request.param not in travis.new_origins:
        pytest.skip('Skipping non-requested origins')
    return request.param


@pytest.fixture(params=modes)
def mode(request):
    """Yield a data mode."""
    if travis and request.param not in travis.new_modes:
        pytest.skip('Skipping non-requested modes')
    return request.param


# Config-based FIXTURES
# ----------------------
@pytest.fixture(scope='session', autouse=True)
def set_config():
    """Set config."""
    config.use_sentry = False
    config.add_github_message = False
    config._traceback = None


@pytest.fixture()
def check_config():
    """Check the config to see if a db is on."""
    return config.db is None


@pytest.fixture(scope='session')
def set_sasurl(loc='local', port=None):
    """Set the sasurl to local or test-utah, and regenerate the urlmap."""
    if not port:
        port = int(os.environ.get('LOCAL_MARVIN_PORT', 5000))
    istest = True if loc == 'utah' else False
    config.switchSasUrl(loc, test=istest, port=port)
    response = Interaction('api/general/getroutemap', request_type='get')
    config.urlmap = response.getRouteMap()


@pytest.fixture(scope='session', autouse=True)
def saslocal():
    """Set sasurl to local."""
    set_sasurl(loc='local')


@pytest.fixture(scope='session')
def urlmap(set_sasurl):
    """Yield the config urlmap."""
    return config.urlmap


@pytest.fixture(scope='session')
def set_release(release):
    """Set the release in the config."""
    config.setMPL(release)


@pytest.fixture(scope='session')
def versions(release):
    """Yield the DRP and DAP versions for a release."""
    drpver, dapver = config.lookUpVersions(release)
    return drpver, dapver


@pytest.fixture(scope='session')
def drpver(versions):
    """Return DRP version."""
    drpver, __ = versions
    return drpver


@pytest.fixture(scope='session')
def dapver(versions):
    """Return DAP version."""
    __, dapver = versions
    return dapver


def set_the_config(release):
    """Set config release without parametrizing.

    Using ``set_release`` combined with ``galaxy`` double parametrizes!"""
    config.setRelease(release)
    config._traceback = None


# DB-based FIXTURES
# -----------------

class DB(object):
    """Object representing aspects of the marvin db.

    Useful for tests needing direct DB access.
    """

    def __init__(self):
        """Initialize with DBs."""
        self._marvindb = marvindb
        self.session = marvindb.session
        self.datadb = marvindb.datadb
        self.sampledb = marvindb.sampledb
        self.dapdb = marvindb.dapdb


@pytest.fixture(scope='session')
def maindb():
    """Yield an instance of the DB object."""
    yield DB()


@pytest.fixture(scope='function')
def db_off():
    """Turn the DB off for a test, and reset it after."""
    config.forceDbOff()
    yield
    config.forceDbOn()


@pytest.fixture(autouse=True)
def db_on():
    """Automatically turn on the DB at collection time."""
    config.forceDbOn()


@pytest.fixture()
def usedb(request):
    ''' fixture for optional turning off the db '''
    if request.param:
        config.forceDbOn()
    else:
        config.forceDbOff()
    return config.db is not None


@pytest.fixture(params=dbs)
def db(request):
    """Turn local db on or off.

    Use this to parametrize over all db options.
    """
    if travis and request.param not in travis.new_dbs:
        pytest.skip('Skipping non-requested dbs')
    if request.param == 'db':
        config.forceDbOn()
    else:
        config.forceDbOff()
    yield config.db is not None
    config.forceDbOn()


@pytest.fixture()
def exporigin(mode, db):
    """Return the expected origins for a given db/mode combo."""
    if mode == 'local' and not db:
        return 'file'
    elif mode == 'local' and db:
        return 'db'
    elif mode == 'remote' and not db:
        return 'api'
    elif mode == 'remote' and db:
        return 'api'
    elif mode == 'auto' and db:
        return 'db'
    elif mode == 'auto' and not db:
        return 'file'


@pytest.fixture()
def expmode(mode, db):
    ''' expected modes for a given db/mode combo '''
    if mode == 'local' and not db:
        return None
    elif mode == 'local' and db:
        return 'local'
    elif mode == 'remote' and not db:
        return 'remote'
    elif mode == 'remote' and db:
        return 'remote'
    elif mode == 'auto' and db:
        return 'local'
    elif mode == 'auto' and not db:
        return 'remote'


# Monkeypatch-based FIXTURES
# --------------------------
@pytest.fixture()
def monkeyconfig(request, monkeypatch):
    """Monkeypatch a variable on the Marvin config.

    Example at line 160 in utils/test_general.
    """
    name, value = request.param
    monkeypatch.setattr(config, name, value=value)


@pytest.fixture()
def monkeymanga(monkeypatch, temp_scratch):
    """Monkeypatch the environ to create a temp SAS dir for reading/writing/downloading.

    Example at line 141 in utils/test_images.
    """
    monkeypatch.setitem(os.environ, 'SAS_BASE_DIR', str(temp_scratch))
    monkeypatch.setitem(os.environ, 'MANGA_SPECTRO_REDUX',
                        str(temp_scratch.join('mangawork/manga/spectro/redux')))
    monkeypatch.setitem(os.environ, 'MANGA_SPECTRO_ANALYSIS',
                        str(temp_scratch.join('mangawork/manga/spectro/analysis')))


# Temp Dir/File-based FIXTURES
# ----------------------------
@pytest.fixture(scope='session')
def temp_scratch(tmpdir_factory):
    """Create a temporary scratch space for reading/writing.

    Use for creating temp dirs and files.

    Example at line 208 in tools/test_query, line 254 in tools/test_results, and
    misc/test_marvin_pickle.
    """
    fn = tmpdir_factory.mktemp('scratch')
    return fn


def tempafile(path, temp_scratch):
    """Return a pytest temporary file given the original file path.

    Example at line 141 in utils/test_images.
    """
    redux = os.getenv('MANGA_SPECTRO_REDUX')
    anal = os.getenv('MANGA_SPECTRO_ANALYSIS')
    endredux = path.partition(redux)[-1]
    endanal = path.partition(anal)[-1]
    end = (endredux or endanal)

    return temp_scratch.join(end)


# Object-based FIXTURES
# ---------------------

class Galaxy(object):
    """An example galaxy for Marvin-tools testing."""

    sasbasedir = os.getenv('SAS_BASE_DIR')
    mangaredux = os.getenv('MANGA_SPECTRO_REDUX')
    mangaanalysis = os.getenv('MANGA_SPECTRO_ANALYSIS')
    dir3d = 'stack'

    def __init__(self, plateifu):
        """Initialize plate and ifu."""
        self.plateifu = plateifu
        self.plate, self.ifu = self.plateifu.split('-')
        self.plate = int(self.plate)

    def set_galaxy_data(self, data_origin=None):
        """Set galaxy properties from the configuration file."""

        if self.plateifu not in galaxy_data:
            return

        data = galaxy_data[self.plateifu]

        for key in data.keys():
            setattr(self, key, data[key])

        # sets specfic data per release
        releasedata = self.releasedata[self.release]
        for key in releasedata.keys():
            setattr(self, key, releasedata[key])

    def set_params(self, bintype=None, template=None, release=None):
        """Set bintype, template, etc."""
        self.release = release
        self.drpver, self.dapver = config.lookUpVersions(self.release)
        self.drpall = 'drpall-{0}.fits'.format(self.drpver)

        default_bintemp = _get_bintemps(self.dapver, default=True)
        default_bin, default_temp = default_bintemp.split('-', 1)

        self.bintype = bintype if bintype is not None else default_bin
        self.template = template if template is not None else default_temp

        self.bintemp = '{0}-{1}'.format(self.bintype, self.template)

        if release == 'MPL-4':
            self.niter = int('{0}{1}'.format(__TEMPLATES_KIN_MPL4__.index(self.template),
                                             __BINTYPES_MPL4__[self.bintype]))
        else:
            self.niter = '*'

        self.access_kwargs = {'plate': self.plate, 'ifu': self.ifu, 'drpver': self.drpver,
                              'dapver': self.dapver, 'dir3d': self.dir3d, 'mpl': self.release,
                              'bintype': self.bintype, 'n': self.niter, 'mode': '*',
                              'daptype': self.bintemp}

    def set_filepaths(self, pathtype='full'):
        """Set the paths for cube, maps, etc."""
        self.path = Path()
        self.imgpath = self.path.__getattribute__(pathtype)('mangaimage', **self.access_kwargs)
        self.cubepath = self.path.__getattribute__(pathtype)('mangacube', **self.access_kwargs)
        self.rsspath = self.path.__getattribute__(pathtype)('mangarss', **self.access_kwargs)

        if self.release == 'MPL-4':
            self.mapspath = self.path.__getattribute__(pathtype)('mangamap', **self.access_kwargs)
            self.modelpath = None
        else:
            __ = self.access_kwargs.pop('mode')
            self.mapspath = self.path.__getattribute__(pathtype)('mangadap5', mode='MAPS',
                                                                 **self.access_kwargs)
            self.modelpath = self.path.__getattribute__(pathtype)('mangadap5', mode='LOGCUBE',
                                                                  **self.access_kwargs)

    def get_location(self, path):
        """Extract the location from the input path."""
        return self.path.location("", full=path)

    def partition_path(self, path):
        """Partition the path into non-redux/analysis parts."""
        endredux = path.partition(self.mangaredux)[-1]
        endanalysis = path.partition(self.mangaanalysis)[-1]
        end = (endredux or endanalysis)
        return end

    def new_path(self, name, newvar):
        ''' Sets a new path with the subsituted name '''
        access_copy = self.access_kwargs.copy()
        access_copy['mode'] = '*'
        access_copy.update(**newvar)

        if name == 'maps':
            access_copy['mode'] = 'MAPS'
            name = 'mangamap' if self.release == 'MPL-4' else 'mangadap5'
        elif name == 'modelcube':
            access_copy['mode'] = 'LOGCUBE'
            name = None if self.release == 'MPL-4' else 'mangadap5'

        path = self.path.full(name, **access_copy) if name else None
        return path


@pytest.fixture(scope='function')
def galaxy(get_params, plateifu):
    """Yield an instance of a Galaxy object for use in tests."""
    release, bintype, template = get_params

    set_the_config(release)
    gal = Galaxy(plateifu=plateifu)
    gal.set_params(bintype=bintype, template=template, release=release)
    gal.set_filepaths()
    gal.set_galaxy_data()

    yield gal
    gal = None


@pytest.fixture()
def cube(galaxy, exporigin, mode):
    ''' Yield a Marvin Cube based on the expected origin combo of (mode+db).
        Fixture tests 6 cube origins from (mode+db) combos [file, db and api]
    '''
    if exporigin == 'file':
        c = Cube(filename=galaxy.cubepath, release=galaxy.release, mode=mode)
    else:
        c = Cube(plateifu=galaxy.plateifu, release=galaxy.release, mode=mode)
    c.exporigin = exporigin
    yield c
    c = None


@pytest.fixture()
def modelcube(galaxy, exporigin, mode):
    ''' Yield a Marvin ModelCube based on the expected origin combo of (mode+db).
        Fixture tests 6 modelcube origins from (mode+db) combos [file, db and api]
    '''
    if exporigin == 'file':
        mc = ModelCube(filename=galaxy.modelpath, release=galaxy.release, mode=mode)
    else:
        mc = ModelCube(plateifu=galaxy.plateifu, release=galaxy.release, mode=mode)
    mc.exporigin = exporigin
    yield mc
    mc = None


@pytest.fixture()
def maps(galaxy, exporigin, mode):
    ''' Yield a Marvin Maps based on the expected origin combo of (mode+db).
        Fixture tests 6 cube origins from (mode+db) combos [file, db and api]
    '''
    if exporigin == 'file':
        m = Maps(filename=galaxy.mapspath, release=galaxy.release, mode=mode)
    else:
        m = Maps(plateifu=galaxy.plateifu, release=galaxy.release, mode=mode)
    m.exporigin = exporigin
    yield m
    m = None


@pytest.fixture()
def query(request, release, mode, db):
    ''' Yields a Query that loops over all modes and db options '''
    data = query_data[release]
    set_the_config(release)
    if mode == 'local' and not db:
        pytest.skip('cannot use queries in local mode without a db')
    searchfilter = request.param if hasattr(request, 'param') else None
    q = Query(searchfilter=searchfilter, mode=mode, release=release)
    q.expdata = data
    yield q
    config.forceDbOn()
    q = None


# @pytest.fixture(autouse=True)
# def skipall():
#     pytest.skip('skipping everything')





# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-04-06 22:36:32
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-04-06 22:37:16

from __future__ import print_function, division, absolute_import
import multiprocessing
import time
import requests
import pytest


class LiveServer(object):
    """The helper class uses to manage live server. Handles creation and
    stopping application in a separate process.
    :param app: The application to run.
    :param port: The port to run application.
    """

    def __init__(self, app, port):
        self.app = app
        self.port = port
        self._process = None

    def start(self):
        """Start application in a separate process."""
        def worker(app, port):
            app.run(port=port, use_reloader=False, threaded=True)
        self._process = multiprocessing.Process(
            target=worker,
            args=(self.app, self.port)
        )
        self._process.start()

        # We must wait for the server to start listening with a maximum
        # timeout of 5 seconds.
        timeout = 5
        while timeout > 0:
            time.sleep(1)
            try:
                requests.get(self.url())
                timeout = 0
            except:
                timeout -= 1

    def url(self, url=''):
        """Returns the complete url based on server options."""
        return 'http://localhost:%d%s' % (self.port, url)

    def stop(self):
        """Stop application process."""
        if self._process:
            self._process.terminate()

    def __repr__(self):
        return '<LiveServer listening at %s>' % self.url()


@pytest.fixture(scope='session')
def live_server(request, app):
    """Run application in a separate process.
    When the ``live_server`` fixture is applyed, the ``url_for`` function
    works as expected::
        def test_server_is_up_and_running(live_server):
            index_url = url_for('index', _external=True)
            assert index_url == 'http://localhost:5000/'
            res = urllib2.urlopen(index_url)
            assert res.code == 200
    """
    # Bind to an open port
    port = app.config.get('LIVESERVER_PORT', 5000)

    # Explicitly set application ``SERVER_NAME`` for test suite
    # and restore original value on test teardown.
    server_name = app.config['SERVER_NAME'] or 'localhost'
    app.config['SERVER_NAME'] = '{0}:{1}'.format(server_name, port)

    server = LiveServer(app, port)
    if request.config.getvalue('start_live_server'):
        server.start()

    yield server

    # teardown
    server.stop()



"""
SQLAlchemy-boolean-search
=========================
SQLAlchemy-boolean-search translates a boolean search string such as::

    "field1=*something* and not (field2==1 or parent.field3<=10.0)"

into its corresponding SQLAlchemy query filter::

    and_(DataModel.field1.ilike('%something%'),
         not_(or_(DataModel.field2.__eq__(2),
                  DataModel.parent.field3.__le__(10.0))))

Relationship field names such as 'parent.grandparent.name' are accepted.

The code is stable, is used in production, and enjoys a test coverage of 100%.

Documentation
-------------
`SQLAlchemy-boolean-search Documentation <http://sqlalchemy-boolean-search.readthedocs.org/>`_

Authors
-------
* Ling Thio - ling.thio [at] gmail.com

"""

from setuptools import setup

setup(
    name='SQLAlchemy-boolean-search',
    version='0.1.1',
    url='http://github.com/lingthio/SQLAlchemy-boolean-search',
    license='BSD License',
    author='Ling Thio',
    author_email='ling.thio@gmail.com',
    description='Boolean search expression parser for SQLAlchemy',
    long_description=__doc__,
    keywords='Boolean search Flask SQLAlchemy',
    py_modules=['sqlalchemy_boolean_search'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask',
        'Flask-SQLAlchemy',
        'pyparsing',
    ],
    test_suite="flask_user.tests.run_tests",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

import os
import pytest
from flask import Flask
from flask_sqlalchemy import SQLAlchemy


the_app = Flask(__name__)  # The WSGI compliant web application object
the_db = SQLAlchemy(the_app)  # Setup Flask-SQLAlchemy

the_app.config.update(
    SECRET_KEY='KeepThisSecret',
    SQLALCHEMY_DATABASE_URI='sqlite:///app.sqlite',
)


@pytest.fixture(scope='session')
def app(request):
    # Establish an application context before running the tests.
    ctx = the_app.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)
    return the_app


@pytest.fixture(scope='session')
def db(app, request):
    """Session-wide test database."""

    def teardown():
        the_db.drop_all()

    the_db.create_all()

    request.addfinalizer(teardown)
    return the_db


@pytest.fixture(scope='function')
def session(db, request):
    """Creates a new database session for a test."""
    connection = db.engine.connect()
    transaction = connection.begin()

    options = dict(bind=connection, binds={})
    session = db.create_scoped_session(options=options)

    db.session = session

    def teardown():
        transaction.rollback()
        connection.close()
        session.remove()

    request.addfinalizer(teardown)
    return session






import re
import functools
from copy import deepcopy
from flask import request
from brain.api.base import processRequest
from marvin import config
from marvin.core.exceptions import MarvinError
from marvin.utils.dap.datamodel import dap_datamodel as dm
from marvin.tools.maps import _get_bintemps
from webargs import fields, validate, ValidationError
from webargs.flaskparser import use_args, use_kwargs, parser


def plate_in_range(val):
    if int(val) < 6500:
        raise ValidationError('Plateid must be > 6500')


# List of global View arguments across all API routes
viewargs = {'name': fields.String(required=True, location='view_args', validate=[validate.Length(min=4),
                                  validate.Regexp('^[0-9-]*$')]),
            'galid': fields.String(required=True, location='view_args', validate=[validate.Length(min=4),
                                   validate.Regexp('^[0-9-]*$')]),
            'bintype': fields.String(required=True, location='view_args'),
            'template_kin': fields.String(required=True, location='view_args'),
            'property_name': fields.String(required=True, location='view_args'),
            'channel': fields.String(required=True, location='view_args'),
            'binid': fields.Integer(required=True, location='view_args', validate=validate.Range(min=-1, max=5800)),
            'plateid': fields.String(required=True, location='view_args', validate=[validate.Length(min=4, max=5),
                                     plate_in_range]),
            'x': fields.Integer(required=True, location='view_args', validate=validate.Range(min=0, max=100)),
            'y': fields.Integer(required=True, location='view_args', validate=validate.Range(min=0, max=100)),
            'mangaid': fields.String(required=True, location='view_args', validate=validate.Length(min=4, max=20)),
            'paramdisplay': fields.String(required=True, location='view_args', validate=validate.OneOf(['all', 'best'])),
            'cube_extension': fields.String(required=True, location='view_args',
                                            validate=validate.OneOf(['flux', 'ivar', 'mask']))
            }

# List of all form parameters that are needed in all the API routes
# allow_none = True allows for the parameter to be non-existent when required=False and missing is not set
# setting missing = None by itself also works except when the parameter is also required
# (i.e. required=True + missing=None does not trigger the "required" validation error when it should)
params = {'query': {'searchfilter': fields.String(allow_none=True),
                    'paramdisplay': fields.String(allow_none=True, validate=validate.OneOf(['all', 'best'])),
                    'task': fields.String(allow_none=True, validate=validate.OneOf(['clean', 'getprocs'])),
                    'start': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'end': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'offset': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'limit': fields.Integer(missing=100, validate=validate.Range(max=50000)),
                    'sort': fields.String(allow_none=True),
                    'order': fields.String(missing='asc', validate=validate.OneOf(['asc', 'desc'])),
                    'rettype': fields.String(allow_none=True, validate=validate.OneOf(['cube', 'spaxel', 'maps', 'rss', 'modelcube'])),
                    'params': fields.DelimitedList(fields.String(), allow_none=True)
                    },
          'search': {'searchbox': fields.String(required=True),
                     'parambox': fields.DelimitedList(fields.String(), allow_none=True)
                     },
          'index': {'galid': fields.String(allow_none=True, validate=[validate.Length(min=4), validate.Regexp('^[0-9-]*$')]),
                    'mplselect': fields.String(allow_none=True, validate=validate.Regexp('MPL-[1-9]'))
                    },
          'galaxy': {'plateifu': fields.String(allow_none=True, validate=validate.Length(min=8, max=11)),
                     'toggleon': fields.String(allow_none=True, validate=validate.OneOf(['true', 'false'])),
                     'image': fields.Url(allow_none=True),
                     'imheight': fields.Integer(allow_none=True, validate=validate.Range(min=0, max=1000)),
                     'imwidth': fields.Integer(allow_none=True, validate=validate.Range(min=0, max=1000)),
                     'type': fields.String(allow_none=True, validate=validate.OneOf(['optical', 'heatmap'])),
                     'x': fields.String(allow_none=True),
                     'y': fields.String(allow_none=True),
                     'mousecoords[]': fields.List(fields.String(), allow_none=True),
                     'bintemp': fields.String(allow_none=True),
                     'params[]': fields.List(fields.String(), allow_none=True)
                     }
          }


# Add a custom Flask session location handler
@parser.location_handler('session')
def parse_session(req, name, field):
    from flask import session as current_session
    value = current_session.get(name, None)
    return value


class ArgValidator(object):
    ''' Web/API Argument validator '''

    def __init__(self, urlmap={}):
        self.release = None
        self.endpoint = None
        self.dapver = None
        self.urlmap = urlmap
        self.base_args = {'release': fields.String(required=True,
                          validate=validate.Regexp('MPL-[1-9]'))}
        self.use_params = None
        self._required = None
        self._setmissing = None
        self._main_kwargs = {}
        self.final_args = {}
        self.final_args.update(self.base_args)

        self._parser = parser
        self.use_args = use_args
        self.use_kwargs = use_kwargs

    def _reset_final_args(self):
        ''' Resets the final args dict '''
        self.final_args = {}
        self.final_args.update(self.base_args)

    def _get_url(self):
        ''' Retrieve the URL route from the map based on the request endpoint '''
        blue, end = self.endpoint.split('.', 1)
        url = self.urlmap[blue][end]['url']
        # if the blueprint is not api, add/remove session option location
        if blue == 'api':
            if 'session' in parser.locations:
                pl = list(parser.locations)
                pl.remove('session')
                parser.locations = tuple(pl)
        else:
            if 'session' not in parser.locations:
                parser.locations += ('session', )

        return url

    def _extract_view_args(self):
        ''' Extract any view argument parameters contained in the URL route '''
        url = self._get_url()
        url_viewargs = re.findall(r'{(.*?)}', url)
        return url_viewargs

    def _add_param_args(self):
        ''' Adds all appropriate form arguments into dictionary for validation '''

        # get the url
        url = self._get_url()

        # check list or not
        self.use_params = [self.use_params] if not isinstance(self.use_params, (list, tuple)) else self.use_params

        for local_param in self.use_params:
            if local_param in params:
                # update param validation
                if self._required or self._setmissing:
                    newparams = self.update_param_validation(local_param)
                else:
                    newparams = deepcopy(params)

                # add to params final args
                self.final_args.update(newparams[local_param])

    def _set_params_required(self, subset):
        ''' Set the param validation required parameter '''

        # make list or not
        self._required = [self._required] if not isinstance(self._required, (list, tuple)) else self._required

        # update the required attribute
        for req_param in self._required:
            if req_param in subset.keys():
                subset[req_param].required = True
                subset[req_param].allow_none = False

            if req_param == 'bintemp':
                bintemps = self._get_bin_temps()
                subset[req_param].validate = validate.OneOf(bintemps)
                subset[req_param].validators.append(validate.OneOf(bintemps))

        return subset

    def _set_params_missing(self, subset):
        ''' Set the param validation missing parameter '''

        for miss_field in subset.values():
            miss_field.missing = None

        return subset

    def update_param_validation(self, name):
        ''' Update the validation of form params '''

        # deep copy the global parameter dict
        newparams = deepcopy(params)
        subset = newparams[name]

        # Set required params
        if self._required:
            subset = self._set_params_required(subset)

        # Set missing params
        if self._setmissing:
            subset = self._set_params_missing(subset)

        # return the new params
        newparams[name] = subset
        return newparams

    def _add_view_args(self):
        ''' Adds all appropriate View arguments into dictionary for validation '''
        local_viewargs = self._extract_view_args()

        # check if any local_view args need new validation
        props = ['bintype', 'template_kin', 'property_name', 'channel']
        ismatch = set(local_viewargs) & set(props)
        if ismatch:
            self.update_view_validation()

        # add only the local view args to the final arg list
        if local_viewargs:
            for varg in local_viewargs:
                self.final_args.update({varg: viewargs[varg]})

    def _update_viewarg(self, name, choices):
        ''' Updates the global View arguments validator '''
        viewargs[name] = fields.String(required=True, location='view_args', validate=validate.OneOf(choices))
        # viewargs[name].validate = validate.OneOf(choices)

    def _get_bin_temps(self):
        ''' Gets the bintemps for a given release '''
        bintemps = _get_bintemps(self.dapver)
        return bintemps

    def update_view_validation(self):
        ''' Update the validation of DAP MPL specific names based on the datamodel '''

        # update all the dap datamodel specific options
        bintemps = self._get_bin_temps()
        bintypes = list(set([b.split('-', 1)[0] for b in bintemps]))
        temps = list(set([b.split('-', 1)[1] for b in bintemps]))
        properties = dm[self.dapver].list_names()
        channels = list(set(sum([i.channels for i in dm[self.dapver] if i.channels is not None], []))) + ['None']

        # update the global viewargs for each property
        propfields = {'bintype': bintypes, 'template_kin': temps, 'property_name': properties,
                      'channel': channels}
        for key, val in propfields.items():
            self._update_viewarg(key, val)

    def create_args(self):
        ''' Build the final argument list for webargs validation '''

        # get the dapver
        drpver, self.dapver = config.lookUpVersions(self.release)

        # reset the final args
        self._reset_final_args()

        # add view args to the list
        self._add_view_args()

        # add form param args to the list
        if self.use_params:
            self._add_param_args()

    def _pop_kwargs(self, **kwargs):
        ''' Pop all non webargs kwargs out of the main kwarg dict '''

        webargs_kwargs = ['req', 'locations', 'as_kwargs', 'validate']
        tempkwargs = kwargs.copy()
        for key in kwargs:
            if key not in webargs_kwargs:
                tmp = tempkwargs.pop(key, None)
        self._main_kwargs = tempkwargs

    def _check_mainkwargs(self, **kwargs):
        self.use_params = kwargs.pop('use_params', None)
        self._required = kwargs.pop('required', None)
        self._setmissing = kwargs.pop('set_missing', None)

    def _get_release_endpoint(self, view):
        ''' get the release and endpoint if you can '''
        self.release = view._release
        self.endpoint = view._endpoint

    def check_args(self, **mainkwargs):
        ''' Checks the input view and parameter arguments for validation using webargs

        This is a decorator and modifies the standard webargs.flaskparser use_args decorator

        '''

        # self note:  nothing can go out here since the decorator is called a lot of overwrites shit

        # decorator used to grab the release and endpoint of the route
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # args and kwargs here are the view function args and kwargs
                self._get_release_endpoint(args[0])
                # self.release = args[0]._release
                # self.endpoint = args[0]._endpoint
                # check the kwargs for any parameters
                self._check_mainkwargs(**mainkwargs)
                # create the arguments dictionary
                self.create_args()
                # pop all the kwargs
                self._pop_kwargs(**mainkwargs)
                # pass into webargs use_args (use_args is a decorator in itself)
                newfunc = self.use_args(self.final_args, **self._main_kwargs)(func)
                return newfunc(*args, **kwargs)
            return wrapper

        return decorator

    def check_kwargs(self, **kwargs):
        kwargs['as_kwargs'] = True
        return self.check_args(**kwargs)

    def check_release(self, **kwargs):
        ''' Checks only the release '''
        return self.use_kwargs(self.base_args)

    def list_params(self, param_type=None):
        ''' List the globally defined parameters for validation
        '''
        total = {'viewargs': viewargs, 'params': params}

        if param_type == 'viewargs':
            return total['viewargs']
        elif param_type == 'params':
            return total['params']
        else:
            return total

    def manual_parse(self, view, req, **mainkwargs):
        ''' Manually parse the args '''
        # args = parser.parse(user_args, request)
        self._get_release_endpoint(view)
        url = self._get_url()
        self._check_mainkwargs(**mainkwargs)
        self.create_args()
        self._pop_kwargs(**mainkwargs)
        newargs = parser.parse(self.final_args, req, force_all=True, **self._main_kwargs)

        # see if we make it a multidict
        makemulti = mainkwargs.get('makemulti', None)
        if makemulti:
            from werkzeug.datastructures import ImmutableMultiDict
            newargs = ImmutableMultiDict(newargs.copy())
        return newargs
